import logging
import numpy as np
import numpy.linalg as la
from .bch import Reader

import scipy.optimize as so
import scipy.interpolate as ip


logger = logging.getLogger(__name__)

KTH = 0.0039  # temperature coefficient of resistance
TREF = 20.0  # reference temperature of resistance


def kskinl(xi, nl):
    xi2 = 2*xi
    nl2 = nl*nl
    return 3 / (nl2*xi2)*(np.sinh(xi2) - np.sin(xi2)) / \
        (np.cosh(xi2)-np.cos(xi2)) + \
        ((nl2-1)/(nl2*xi)*(np.sinh(xi)+np.sin(xi)) /
            (np.cosh(xi)+np.cos(xi)))


def kskinr(xi, nl):
    xi2 = 2*xi
    nl2 = nl*nl
    return xi*((np.sinh(xi2)+np.sin(xi2))/(np.cosh(xi2)-np.cos(xi2))) + \
        ((nl2-1) / 3 * xi2*((np.sinh(xi)-np.sin(xi)) /
                            (np.cosh(xi)+np.cos(xi))))


def resistance(r0, w, temp, zeta, gam, nh):
    xi = zeta*np.sqrt(abs(w)/(2*np.pi)/(50*(1+KTH*(temp-TREF))))
    if np.isscalar(xi):
        if xi < 1e-12:
            k = 1
        else:
            k = (gam + kskinr(xi, nh)) / (1. + gam)
    else:
        k = np.ones(np.asarray(w).shape)
        k[xi > 1e-12] = (gam + kskinr(xi[xi > 1e-12], nh)) / (1. + gam)
    return r0*(1.+KTH*(temp - TREF))*k


def mesh(x, y):
    """return the combined vectors x and y
    """
    size = np.asarray(x).size, np.asarray(y).size
    return (np.repeat(x, size[1]),
            np.tile(y, size[0]))


def K(d):
    """space phasor transformation matrix
    (Inverse Park Transformation) T-1 * dq
    arguments:
      d: rotation angle

    returns transformation matrix
    """
    return np.array((
        (-np.cos(d), np.sin(d)),
        (-np.cos(d-2*np.pi/3), np.sin(d-2*np.pi/3)),
        (-np.cos(d+2*np.pi/3), np.sin(d+2*np.pi/3))))


def T(d):
    """space phasor transformation matrix
    (Park Transformation) T * abc
    arguments:
      d: rotation angle

    returns transformation matrix
    """
    return np.array((
        (-np.cos(d), -np.cos(d-2*np.pi/3), -np.cos(d+2*np.pi/3)),
        (np.sin(d), np.sin(d-2*np.pi/3), np.sin(d+2*np.pi/3))))/3*2


def invpark(a, q, d):
    """ convert a dq vector to the abc reference frame
    (inverse park transformation)

    Args:
        a: rotation angle
        d: value in direct axis
        q: value in quadrature axis
    """
    if np.isscalar(a) and np.isscalar(q) and np.isscalar(d):
        return np.dot(K(a), (q, d))
    if np.isscalar(q) and np.isscalar(d):
        return np.array([K(x).dot((q, d)) for x in a]).T
    return np.array([K(x).dot((y, z)) for x, y, z in zip(a, d, q)]).T


def betai1(iq, id):
    """return beta and amplitude of dq currents"""
    return (np.arctan2(id, iq),
            la.norm((id, iq), axis=0)/np.sqrt(2.0))


def iqd(beta, i1):
    """return qd currents of beta and amplitude"""
    return np.sqrt(2.0)*i1*np.array([np.cos(beta),
                                     np.sin(beta)])


def puconv(dqpar, p, NR, UR, IR):
    """convert dqpar to per unit
    arguments:
    dqpar: dict from ld-iq or psid-psiq identification
    p: pole pairs
    NR: ref speed in 1/s
    UR: ref voltage per phase in V
    IR: ref current per phase in A
    """
    WR = 2*np.pi*NR*p
    PSIR = UR/WR
    SR = 3*UR*IR
    if 'beta' in dqpar:
        dqp = dict(beta=dqpar['beta'], losses=dict())
        dqp['i1'] = np.array(dqpar['i1'])/IR
    elif 'iq' in dqpar:
        dqp = dict(iq=np.array(dqpar['iq)'])/IR*np.sqrt(2), losses=dict())
        dqp['id'] = np.array(dqpar['id'])/IR*np.sqrt(2)
    else:
        raise ValueError('invalid dqpar')
    for k in 'psid', 'psiq':
        dqp[k] = np.array(dqpar[k])/PSIR
    if 'losses' in dqpar:
        for k in ('magnet', 'styoke_hyst', 'styoke_eddy',
                  'stteeth_hyst', 'stteeth_eddy', 'rotor_hyst', 'rotor_eddy'):
            dqp['losses'][k] = np.array(dqpar['losses'][k])/SR
        dqp['losses']['speed'] = p*dqpar['losses']['speed']/WR
        dqp['losses']['ef'] = dqpar['losses']['ef']
        dqp['losses']['hf'] = dqpar['losses']['hf']
    return dqp


def __scale_losses(losses, wdg, lfe):
    if losses:
        l = {k: wdg*lfe*np.array(losses[k]) for k in (
            'styoke_hyst', 'styoke_eddy',
            'stteeth_hyst', 'stteeth_eddy',
            'rotor_hyst', 'rotor_eddy',
            'magnet')}
        l['speed'] = losses['speed']
        return l
    return {}


def dqpar_interpol(xfit, dqpars, ipkey='temperature'):
    """return interpolated parameters at temperature or exc_current

    Arguments:
      xfit -- temperature or exc_current to fit dqpars
      dqpars -- list of dict with id, iq (or i1, beta), Psid and Psiq values
      ipkey -- key (string) to interpolate
    """
    # check current range
    ckeys = (('i1', 'beta'), ('id', 'iq'))
    dqtype = 0
    fpip = {k: dqpars[0][k] for k in ckeys[dqtype]}
    fpip['losses'] = dict()
    for k in ckeys[dqtype]:
        curr = np.array([f[k] for f in dqpars], dtype=object)
        shape = curr.shape
        if curr.shape != (len(dqpars), len(curr[0])):
            raise ValueError("current range conflict")
        curr = curr.astype(float)
        if not np.array([np.allclose(curr[0], c)
                         for c in curr[1:]]).all():
            raise ValueError("current range conflict")

    try:
        speed = np.array([d['losses']['speed'] for d in dqpars])
        if (np.max(speed) - np.min(speed))/np.mean(speed) > 1e-3:
            raise ValueError("losses: speed conflict")
    except KeyError:
        pass

    sorted_dqpars = sorted(dqpars, key=lambda d: d[ipkey])
    x = [f[ipkey] for f in sorted_dqpars]
    for k in ('psid', 'psiq'):
        m = np.array([f[k] for f in sorted_dqpars]).T
        if len(x) > 2:
            fpip[k] = np.array(
                [[ip.UnivariateSpline(x, y, k=2)(xfit)
                  for y in row] for row in m]).T
        else:
            fpip[k] = ip.interp1d(
                x, m, fill_value='extrapolate')(xfit).T
    try:
        for k in ('styoke_hyst', 'stteeth_hyst',
                  'styoke_eddy', 'stteeth_eddy',
                  'rotor_hyst', 'rotor_eddy',
                  'magnet'):
            m = np.array([f['losses'][k] for f in sorted_dqpars]).T
            if len(x) > 2:
                fpip['losses'][k] = np.array(
                    [[ip.UnivariateSpline(x, y, k=2)(xfit)
                      for y in row] for row in m]).T
            else:
                fpip['losses'][k] = ip.interp1d(
                    x, m, fill_value='extrapolate')(xfit).T
            fpip['losses']['speed'] = dqpars[0]['losses']['speed']
            for f in ('hf', 'ef'):
                if f in dqpars[0]['losses']:
                    fpip['losses'][f] = dqpars[0]['losses'][f]
    except KeyError:
        pass
    return x, fpip


def wdg_resistance(w1, l, d, sigma=56e3):
    """return winding resistance
    arguments:
    w1: number of turns
    l: wire length of one turn
    d: wire diameter m^2
    sigma: conductivity of wire material
    """
    a = np.pi*d**2/4
    return w1*l/sigma/a


def create(bch, r1, ls, lfe=1, wdg=1):
    """create PmRelMachine from BCH

    Arguments:
      bch: BchReader or Erg object
      r1: winding resistance
      ls: winding leakage
      lfe: scale factor length
      wdg: scale factor number of windings
"""
    m = 3
    if isinstance(bch, Reader):
        p = bch.machine['p']
        if bch.type.lower().find('psid-psiq-identification') >= 0:
            id = np.array(bch.psidq['id'])/wdg
            iq = np.array(bch.psidq['iq'])/wdg
            psid = wdg*lfe*np.array(bch.psidq['psid'])
            psiq = wdg*lfe*np.array(bch.psidq['psiq'])
            try:
                losses = __scale_losses(bch.psidq['losses'], wdg, lfe)
                losses['ef'] = bch.lossPar.get('ef', [2.0, 2.0])
                losses['eh'] = bch.lossPar.get('eh', [1.0, 1.0])
            except KeyError:
                losses = {}
            return PmRelMachinePsidq(m, p, psid, psiq, r1*lfe*wdg**2,
                                     id, iq, ls*wdg**2, losses=losses)

        if bch.type.lower().find('ld-lq-identification') >= 0:
            beta = bch.ldq['beta']
            i1 = np.array(bch.ldq['i1'])/wdg
            psid = wdg*lfe*np.array(bch.ldq['psid'])
            psiq = wdg*lfe*np.array(bch.ldq['psiq'])
            try:
                losses = __scale_losses(bch.ldq['losses'], wdg, lfe)
                losses['ef'] = bch.lossPar.get('ef', [2.0, 2.0])
                losses['eh'] = bch.lossPar.get('eh', [1.0, 1.0])
            except KeyError:
                losses = {}
            return PmRelMachineLdq(m, p, psid=psid, psiq=psiq,
                                   r1=r1*lfe*wdg**2,
                                   i1=i1, beta=beta, ls=ls*wdg**2,
                                   losses=losses)
        raise ValueError("Unsupported BCH type {}".format(bch.type))
    # must be ERG type:
    p = int(round(np.sqrt(2)*bch['M_sim'][-1][-1]/(
        m*bch['Psi_d'][-1][-1] * bch['i1'][-1])))

    return PmRelMachineLdq(m, p, r1=r1*lfe*wdg**2,
                           beta=bch['beta'], i1=np.array(bch['i1'])/wdg,
                           psid=wdg*lfe*np.array(bch['Psi_d'])/np.sqrt(2),
                           psiq=wdg*lfe*np.array(bch['Psi_q'])/np.sqrt(2),
                           ls=ls*wdg**2)


class PmRelMachine(object):
    """Abstract base class for PmRelMachines

    Args:
        m: number of winding phases
        p: number of pole pairs
        r1: stator winding resistance (in Ohm)
        ls: leakage inductance in H
    """

    def __init__(self, m, p, r1, ls, **kwargs):
        self.p = p
        self.m = m
        self.r1 = r1
        self.ls = ls
        self.io = (1, -1)
        self.fo = 50.0
        self.tcu1 = 20
        self.zeta1 = 0.2
        self.gam = 0.7
        self.kh = 2
        for k in kwargs.keys():
            setattr(self, k, kwargs[k])

        self.plexp = {'styoke_hyst': 1.0,
                      'stteeth_hyst': 1.0,
                      'styoke_eddy': 2.0,
                      'stteeth_eddy': 2.0,
                      'rotor_hyst': 1.0,
                      'rotor_eddy': 2.0}
        self._losses = {k: lambda x, y: 0 for k in (
            'styoke_hyst', 'stteeth_hyst',
            'styoke_eddy', 'stteeth_eddy',
            'rotor_hyst', 'rotor_eddy',
            'magnet')}

    def rstat(self, w):
        """stator resistance"""
        return resistance(self.r1, w, self.tcu1, self.zeta1,
                          self.gam, self.kh)

    def torque_iqd(self, iq, id):
        "torque at q-d-current"
        psid, psiq = self.psi(iq, id)
        tq = self.m*self.p/2*(psid*iq - psiq*id)
        return tq

    def iqd_torque(self, torque):
        """return minimum d-q-current for torque"""
        if np.abs(torque) < 1e-2:
            return (0, 0)
        res = so.minimize(lambda iqd: la.norm(iqd), self.io, method='SLSQP',
                          constraints=({'type': 'eq',
                                        'fun': lambda iqd:
                                        self.torque_iqd(*iqd) - torque}))
        if not res.success:
            raise ValueError(f'Torque {torque} out of current range')
        return res.x

    def uqd(self, w1, iq, id):
        """return uq, ud of frequency w1 and d-q current"""
        psid, psiq = self.psi(iq, id)
        uqd = (self.r1*iq + w1*(self.ls*id + psid),
               self.r1*id - w1*(self.ls*iq + psiq))
        logger.debug('beta i1 %s u1 %f', betai1(iq, id), la.norm(uqd))
        return uqd

    def w1_umax(self, u, iq, id):
        """return frequency w1 at given voltage u and id, iq current

        Keyword arguments:
        u -- the maximum voltage (RMS)
        iq, id -- the d-q currents"""
        w10 = np.sqrt(2)*u/la.norm(self.psi(iq, id))
        return so.fsolve(lambda w1:
                         la.norm(self.uqd(w1, iq, id))-u*np.sqrt(2), w10)[0]

    def w1_u(self, u, iq, id):
        """return frequency w1 at given voltage u and id, iq current
        (obsolete, use w1_umax)"""
        return self.w1_umax(u, iq, id)

    def w1max(self, u, iq, id):
        """return max frequency w1 at given voltage u and d-q current
        (obsolete, use w1_umax)"""
        return self.w1_umax(u, iq, id)

    def w2_imax_umax(self, imax, umax, maxtorque=True):
        """return frequency at max current and max voltage"""
        w, info, ier, mesg = so.fsolve(lambda x: np.linalg.norm(
            self.uqd(x, *iqd(-np.pi/2, imax))) - umax*np.sqrt(2),
            np.sqrt(2)*umax/la.norm(self.psi(*self.io)),
            full_output=True)
        if ier == 1:
            return w[0]

        logger.warn("w2_imax_umax ier=%d imax %f", ier, imax)
        raise ValueError("w2_imax_umax {} imax {}".format(mesg, imax))

    def beta_u(self, w1, u, i1):
        "beta at given frequency, voltage and current"
        return so.fsolve(lambda b:
                         la.norm(self.uqd(w1, *(iqd(b, i1))))-u*np.sqrt(2),
                         np.arctan2(self.io[1], self.io[0]))[0]

    def iq_u(self, w1, u, id):
        "iq at given frequency, voltage and id current"
        iq0 = max(self.io[0]/4, id*np.tan(self.betarange[0]))
        return so.fsolve(lambda iq:
                         la.norm(self.uqd(w1, iq, id))-u*np.sqrt(2),
                         iq0)[0]

    def iqd_uqd(self, w1, uq, ud):
        "return iq, id current at given frequency, voltage"
        return so.fsolve(lambda iqd:
                         np.array((uq, ud)) - self.uqd(w1, *iqd),
                         (0, self.io[1]))

    def i1_torque(self, torque, beta):
        "return i1 current with given torque and beta"
        i1, info, ier, mesg = so.fsolve(
            lambda i1: self.torque_iqd(*iqd(beta, i1))-torque,
            self.io[0],
            full_output=True)
        if ier == 1:
            return i1
        raise ValueError("no solution found for torque {}, beta {}".format(
            torque, beta))

    def i1_voltage(self, w1, u1, beta):
        "return i1 current with given w1, u1 and beta"
        i1, info, ier, mesg = so.fsolve(
            lambda i1: la.norm(self.uqd(w1, *iqd(beta, i1)))-np.sqrt(2)*u1,
            la.norm(self.io),
            full_output=True)
        if ier == 1:
            return i1
        raise ValueError("{} for w1 {}, u1 {}, beta {}".format(
            mesg, w1, u1, beta))

    def id_torque(self, torque, iq):
        "return d current with given torque and d-current"
        id0 = min(self.io[1]/4, iq/np.tan(self.betarange[0]))
        return so.fsolve(lambda id: self.torque_iqd(iq, id)-torque, id0)[0]

    def iqd_torque_umax(self, torque, w1, u1max):
        "return d-q current and torque at stator frequency and max voltage"
        iq, id = self.iqd_torque(torque)
        # check voltage
        if la.norm(self.uqd(w1, iq, id)) <= u1max*np.sqrt(2):
            return (iq, id, torque)
        # decrease psi (flux weakening mode), let i1 == i1max
        iqd, info, ier, mesg = so.fsolve(
            lambda iqd: (la.norm(self.uqd(w1, *iqd)) - u1max*np.sqrt(2),
                         self.torque_iqd(*iqd) - torque),
            (iq, id),
            full_output=True)
        if ier != 1:  # didn't converge
            return self.mtpv(w1, u1max, betai1(iq, id)[1],
                             maxtorque=torque > 0)
        return iqd[0], iqd[1], self.torque_iqd(iq, id)

    def iqd_torque_imax_umax(self, torque, n, umax):
        """return iq, id, torque for constant torque or field weakening"""
        iq, id = self.iqd_torque(torque)
        w1 = 2*np.pi*n*self.p
        # Constant torque range
        if np.linalg.norm(self.uqd(w1, iq, id)) <= umax*np.sqrt(2):
            return (iq, id, torque)
        # Field weaking range
        imax = betai1(iq, id)[1]
        iq, id = self.iqd_imax_umax(imax, w1, umax, maxtorque=torque > 0)
        return iq, id, self.torque_iqd(iq, id)

    def iqd_imax_umax(self, i1max, w1, u1max, maxtorque=True):
        """return d-q current at stator frequency and max voltage
        and max current (for motor operation if maxtorque else generator operation)"""

        beta0 = max(
            self.betarange[0],
            -0.7*np.pi/2 if maxtorque else -1.4*np.pi/2)

        beta, info, ier, mesg = so.fsolve(
            lambda b: la.norm(
                self.uqd(w1, *iqd(b, i1max))) - u1max*np.sqrt(2),
            beta0,
            full_output=True)
        if ier == 1:
            return iqd(beta[0], i1max)

        return self.mtpv(w1, u1max, i1max, maxtorque)[:2]

#        raise ValueError(
#            "no solution found for imax {}, w1 {}, u1max {}".format(
#                i1max, w1, u1max))

    def mtpa(self, i1):
        """return iq, id, torque at maximum torque of current i1"""
        sign = -1 if i1 > 0 else 1
        b0 = 0 if i1 > 0 else -np.pi
        bopt, fopt, iter, funcalls, warnflag = so.fmin(
            lambda x: sign*self.torque_iqd(*iqd(x, abs(i1))), b0,
            full_output=True,
            disp=0)
        iq, id = iqd(bopt[0], abs(i1))
        return [iq, id, sign*fopt]

    def mtpv(self, w1, u1, i1max, maxtorque=True):
        """return d-q-current, torque for voltage and frequency
        with maximum (maxtorque=True) or minimum torque """
        sign = -1 if maxtorque else 1
        i0 = (-sign*self.i1range[1]/10, self.i1range[1]/10)

        res = so.minimize(
            lambda iqd: sign*self.torque_iqd(*iqd),
            i0, method='SLSQP',
            constraints=(
                {'type': 'ineq',
                 'fun': lambda iqd:
                 np.sqrt(2)*u1 - la.norm(self.uqd(w1, *iqd))},
                {'type': 'ineq',
                 'fun': lambda iqd:
                 i1max - betai1(*iqd)[1]}))

        return res.x[0], res.x[1], sign*res.fun

    def _set_losspar(self, pfe):
        self.fo = pfe['speed']*self.p
        ef = pfe.get('ef', [2.0, 2.0])
        hf = pfe.get('hf', [1.0, 1.0])
        self.plexp = {'styoke_hyst': hf[0],
                      'stteeth_hyst': hf[0],
                      'styoke_eddy': ef[0],
                      'stteeth_eddy': ef[0],
                      'rotor_hyst': hf[1],
                      'rotor_eddy': ef[1]}
        #                          'magnet'):

    def betai1_plcu(self, i1, w1=0):
        return self.m*self.rstat(w1)*i1**2

    def iqd_plcu(self, iq, id, w1=0):
        return self.m*self.rstat(w1)*(iq**2+id**2)/2

    def iqd_plcu1(self, iq, id, w1):
        return self.iqd_plcu(iq, id, w1)

    def iqd_plcu2(self, iq, id):
        return np.zeros(np.asarray(iq).shape)

    def betai1_losses(self, beta, i1, f):
        return np.sum([self.betai1_plfe1(beta, i1, f),
                       self.betai1_plfe2(beta, i1, f),
                       self.betai1_plmag(beta, i1, f),
                       self.betai1_plcu(i1), 2*np.pi*f], axis=0)

    def iqd_losses(self, iq, id, f):
        return np.sum([self.iqd_plfe1(iq, id, f),
                       self.iqd_plfe2(iq, id, f),
                       self.iqd_plmag(iq, id, f),
                       self.iqd_plcu(iq, id, 2*np.pi*f)], axis=0)

    def characteristics(self, T, n, u1max, nsamples=50):
        """calculate torque speed characteristics.
        return dict with list values of
        id, iq, n, T, ud, uq, u1, i1,
        beta, gamma, phi, cosphi, pmech, n_type

        Keyword arguments:
        T -- the maximum torque or the list of torque values in Nm
        n -- the maximum speed or the list of speed values in 1/s
        u1max -- the maximum voltage in V rms
        nsamples -- (optional) number of speed samples
        """
        r = dict(id=[], iq=[], uq=[], ud=[], u1=[], i1=[], T=[],
                 beta=[], gamma=[], phi=[], cosphi=[], pmech=[], n=[])
        if np.isscalar(T):
            iq, id = self.iqd_torque(T)
            i1max = betai1(iq, id)[1]
            w1 = self.w1_umax(u1max, iq, id)
            w1max = self.w1_umax(u1max, *self.iqdmin(i1max))
            nmax = max(w1, w1max)/2/np.pi/self.p

            n1 = min(w1/2/np.pi/self.p, nmax)
            r['n_type'] = n1
            logger.info("Type speed %f n: %f nmax %f",
                        60*n1, 60*n, 60*nmax)
            try:
                w1 = self.w2_imax_umax(i1max, u1max, maxtorque=T > 0)
                n2 = w1/2/np.pi/self.p
                iqmtpv, idmtpv, tq = self.mtpv(
                    w1, u1max, i1max, maxtorque=T > 0)
                if not self._inrange((iqmtpv, idmtpv)):
                    if n > 0:
                        n2 = min(nmax, n)
                    else:
                        n2 = nmax
                logger.info("n1: %f n2: %f ",
                            60*n1, 60*n2)
            except ValueError:
                if n > 0:
                    n2 = min(nmax, n)
                else:
                    n2 = nmax
            if n > 0:
                speedrange = sorted(
                    list(set([nx for nx in [n1, n2, n] if nx <= n])))
            else:
                speedrange = sorted(list(set([n1, n2])))
            n1 = speedrange[0]
            n3 = speedrange[-1]
            if n2 > n3:
                n2 = n3
            logger.info("Speed intervals %s",
                        [60*nx for nx in speedrange])
            if len(speedrange) > 2:
                nsamples = nsamples - int(speedrange[1]/(n3/nsamples))
                dn = (n3-speedrange[1])/nsamples
            else:
                dn = n3 / nsamples
            nx = n1

            for nx in np.linspace(0, n1, int(n1/dn)):
                r['id'].append(id)
                r['iq'].append(iq)
                r['n'].append(nx)
                r['T'].append(T)

            if n1 < n2:
                for nx in np.linspace(nx+dn/2, n2, int(n2/dn)):
                    w1 = 2*np.pi*nx*self.p
                    iq, id = self.iqd_imax_umax(i1max, w1, u1max,
                                                maxtorque=T > 0)
                    tq = self.torque_iqd(iq, id)
                    r['id'].append(id)
                    r['iq'].append(iq)
                    r['n'].append(nx)
                    r['T'].append(tq)
                    if T > 0 and tq < 0:
                        logger.info("2: n %g T %g i1max %g w1 %g u1 %g",
                                    nx*60, tq, i1max, w1, u1max)
            if n2 < n3:
                for nx in np.linspace(nx+dn/2, n3, int(n3/dn)):
                    w1 = 2*np.pi*nx*self.p
                    try:
                        iq, id, tq = self.mtpv(
                            w1, u1max, i1max, maxtorque=T > 0)
                        if not self._inrange((iq, id)):
                            break
                    except ValueError:
                        logger.warn("ValueError at speed %f", 60*nx)
                        break
                    r['id'].append(id)
                    r['iq'].append(iq)
                    r['n'].append(nx)
                    r['T'].append(tq)

        else:
            for t, nx in zip(T, n):
                w1 = 2*np.pi*nx*self.p
                iq, id, tq = self.iqd_torque_umax(t, w1, u1max)
                r['id'].append(id)
                r['iq'].append(iq)
                r['T'].append(tq)
                r['n'].append(nx)

        for nx, iq, id in zip(r['n'], r['iq'], r['id']):
            w1 = 2*np.pi*nx*self.p
            uq, ud = self.uqd(w1, iq, id)
            r['uq'].append(uq)
            r['ud'].append(ud)
            r['u1'].append(la.norm((ud, uq))/np.sqrt(2.0))
            r['i1'].append(la.norm((id, iq))/np.sqrt(2.0))
            r['beta'].append(np.arctan2(id, iq)/np.pi*180.)
            r['gamma'].append(np.arctan2(ud, uq)/np.pi*180.)

            r['phi'].append(r['beta'][-1] - r['gamma'][-1])
            r['cosphi'].append(np.cos(r['phi'][-1]/180*np.pi))

        pmech = np.array([2*np.pi*nx*tq for nx, tq in zip(r['n'], r['T'])])
        f1 = np.array(r['n'])*self.p
        plfe1 = self.iqd_plfe1(np.array(r['iq']), np.array(r['id']), f1)
        plfe2 = self.iqd_plfe2(np.array(r['iq']), np.array(r['id']), f1)
        plmag = self.iqd_plmag(np.array(r['iq']), np.array(r['id']), f1)
        plfe = plfe1 + plfe2 + plmag
        plcu = self.betai1_plcu(np.array(r['i1']), 2*np.pi*f1)
        pltotal = plfe + plcu
        r['pmech'] = pmech.tolist()
        r['plfe'] = plfe.tolist()
        r['plcu'] = plcu.tolist()
        r['losses'] = pltotal.tolist()
        if pmech.any():
            p1 = pmech + pltotal
            if np.abs(pmech[0]) < 1e-12:
                r['eta'] = [0]
                i = 1
            else:
                r['eta'] = []
                i = 0
            if np.all(abs(p1[i:]) > abs(pmech[i:])):
                r['eta'] += (pmech[i:]/(p1[i:])).tolist()
            else:
                r['eta'] += (p1[i:]/pmech[i:]).tolist()
        return r

    def i1beta_characteristics(self, n_list, i1_list, beta_list, u1max):
        """calculate i1-beta characteristics"""
        r = dict(id=[], iq=[], uq=[], ud=[], u1=[], i1=[], T=[],
                 beta=[], gamma=[], phi=[], cosphi=[], pmech=[], n=[])
        for n, i1, beta in zip(n_list, i1_list, beta_list):
            w1 = 2*np.pi*n*self.p
            beta = beta/180*np.pi
            iq, id = iqd(beta, i1)
            uq, ud = self.uqd(w1, iq, id)
            u1 = la.norm((ud, uq))/np.sqrt(2)
            if u1 > u1max:
                logger.debug("u1 %s > %s", u1, u1max)
                beta = self.beta_u(w1, u1max, i1)
                logger.debug("beta %s", beta*180/np.pi)
                iq, id = iqd(beta, i1)
                logger.debug("beta %s id, %s iq %s", beta*180/np.pi, id, iq)
                uq, ud = self.uqd(w1, iq, id)
                u1 = la.norm((ud, uq))/np.sqrt(2)
                logger.debug("ud %s uq %s --> u1 %s", ud, uq, u1)

            tq = self.torque_iqd(iq, id)

            r['id'].append(id)
            r['iq'].append(iq)

            r['uq'].append(uq)
            r['ud'].append(ud)
            r['u1'].append(u1)
            r['i1'].append(la.norm((id, iq))/np.sqrt(2))
            r['T'].append(tq)
            r['beta'].append(np.arctan2(id, iq)/np.pi*180.)
            r['gamma'].append(np.arctan2(ud, uq)/np.pi*180.)

            r['n'].append(n)
            r['phi'].append(r['beta'][-1]-r['gamma'][-1])
            r['cosphi'].append(np.cos(r['phi'][-1]/180*np.pi))
            r['pmech'].append(w1/self.p*r['T'][-1])

        r['losses'] = self.iqd_losses(
            *iqd(np.array(beta_list)/180*np.pi,
                 np.array(i1_list)),
            np.array(n_list)*self.p).tolist()
        return r

    def _inrange(self, iqd):
        i1 = np.linalg.norm(iqd)/np.sqrt(2)
        iqmin, idmin = self.iqdmin(i1)
        iqmax, idmax = self.iqdmax(i1)
        return iqmin <= iqd[0] <= iqmax and idmin <= iqd[1] <= idmax


class PmRelMachineLdq(PmRelMachine):
    """Standard set of PM machine given by i1,beta parameters:
    p number of pole pairs
    m number of phases
    psim flux in Vs (RMS)
    ld d-inductance in
    lq q-inductance in H
    r1 stator resistance
    ls stator leakage inductance in H
    beta angle i1 vs up in degrees
    i1 current in A (RMS)

    optional keyword args:
    psid D-Flux in Vs (RMS)
    psiq Q-Flux in Vs (RMS)
    """

    def __init__(self,  m, p, psim=[], ld=[], lq=[],
                 r1=0, beta=[], i1=[], ls=0, **kwargs):

        super(self.__class__, self).__init__(m, p, r1, ls)
        self.psid = None
        self.betarange = (-np.pi, np.pi)
        self.i1range = (0, np.inf)
        if np.isscalar(ld):
            self.ld = lambda b, i: ld
            self.psim = lambda b, i: psim
            self.lq = lambda b, i: lq
            logger.debug("ld %s lq %s psim %s", ld, lq, psim)
            return

        if len(ld) == 1:
            try:
                self.io = iqd(min(beta)*np.pi/360, max(i1)/2)
            except:
                self.io = (1, -1)
            self.ld = lambda b, i: ld[0]
            self.psim = lambda b, i: psim[0]
            self.lq = lambda b, i: lq[0]
            logger.debug("ld %s lq %s psim %s", ld, lq, psim)
            return

        beta = np.asarray(beta)/180.0*np.pi
        if np.any(beta[beta > np.pi]):
            beta[beta > np.pi] = beta - 2*np.pi
        self.io = iqd((np.min(beta)+max(beta))/2, np.max(i1)/2)
        kx = ky = 3
        if len(i1) < 4:
            ky = len(i1)-1
        if len(beta) < 4:
            kx = len(beta)-1
        try:
            pfe = kwargs['losses']
            self._set_losspar(pfe)
            self._losses = {k: ip.RectBivariateSpline(
                beta, i1, np.array(pfe[k]),
                kx=kx, ky=ky).ev for k in (
                    'styoke_hyst', 'stteeth_hyst',
                    'styoke_eddy', 'stteeth_eddy',
                    'rotor_hyst', 'rotor_eddy',
                    'magnet')}
        except KeyError:
            logger.warning("loss map missing")
            pass
        if 'psid' in kwargs:
            self.betarange = min(beta), max(beta)
            self.i1range = (0, np.max(i1))
            self.psid = lambda x, y: ip.RectBivariateSpline(
                beta, i1, np.sqrt(2)*np.asarray(kwargs['psid']),
                kx=kx, ky=ky).ev(x, y)
            self.psiq = lambda x, y: ip.RectBivariateSpline(
                beta, i1, np.sqrt(2)*np.asarray(kwargs['psiq']),
                kx=kx, ky=ky).ev(x, y)

            return

        if len(i1) < 4 or len(beta) < 4:
            if len(i1) == len(beta):
                self.ld = lambda x, y: ip.interp2d(beta, i1, ld.T)(x, y)
                self.psim = lambda x, y: ip.interp2d(beta, i1, psim.T)(x, y)
                self.lq = lambda x, y: ip.interp2d(beta, i1, lq.T)(x, y)
                logger.debug("interp2d beta %s i1 %s", beta, i1)
                return
            elif len(i1) == 1:
                self.ld = lambda x, y: ip.InterpolatedUnivariateSpline(
                    beta, ld, k=1)(x)
                self.psim = lambda x, y: ip.InterpolatedUnivariateSpline(
                    beta, psim, k=1)(x)
                self.lq = lambda x, y: ip.InterpolatedUnivariateSpline(
                    beta, lq, k=1)(x)
                logger.debug("interpolatedunivariatespline beta %s", beta)
                return
            if len(beta) == 1:
                self.ld = lambda x, y: ip.InterpolatedUnivariateSpline(
                    i1, ld, k=1)(y)
                self.psim = lambda x, y: ip.InterpolatedUnivariateSpline(
                    i1, ld, k=1)(y)
                self.lq = lambda x, y: ip.InterpolatedUnivariateSpline(
                    i1, lq, k=1)(y)
                logger.debug("interpolatedunivariatespline i1 %s", i1)
                return

            raise ValueError("unsupported array size {}x{}".format(
                len(beta), len(i1)))

        self.betarange = min(beta), max(beta)
        self.i1range = (0, np.max(i1))
        self.ld = lambda x, y: ip.RectBivariateSpline(
            beta, i1, np.asarray(ld)).ev(x, y)
        self.psim = lambda x, y: ip.RectBivariateSpline(
            beta, i1, np.asarray(psim)).ev(x, y)
        self.lq = lambda x, y: ip.RectBivariateSpline(
            beta, i1, np.asarray(lq)).ev(x, y)
        logger.debug("rectbivariatespline beta %s i1 %s", beta, i1)

    def psi(self, iq, id):
        """return psid, psiq of currents iq, id"""
        beta, i1 = betai1(np.asarray(iq), np.asarray(id))
        logger.debug('beta %f (%f, %f) i1 %f %f',
                     beta, self.betarange[0], self.betarange[1],
                     i1, self.i1range[1])
        if (self.betarange[0] <= beta <= self.betarange[1] and
                i1 <= 1.01*self.i1range[1]):
            if self.psid:
                return (self.psid(beta, i1), self.psiq(beta, i1))

            psid = self.ld(beta, i1)*id + np.sqrt(2)*self.psim(beta, i1)
            psiq = self.lq(beta, i1)*iq
            return (psid, psiq)

        return (np.nan, np.nan)

    def iqdmin(self, i1):
        """max iq, min id for given current"""
        if self.betarange[0] <= -np.pi/2 <= self.betarange[1]:
            return iqd(-np.pi/2, i1)
        if self.betarange[1] == 0:
            return iqd(self.betarange[0], i1)
        return iqd(self.betarange[1], i1)

    def iqdmax(self, i1):
        """max iq, min id for given current"""
        if self.betarange[1] == 0:
            return iqd(self.betarange[1], i1)
        return iqd(self.betarange[0], i1)

    def betai1_plfe1(self, beta, i1, f1):
        return np.sum([
            self._losses[k](beta, i1)*(f1/self.fo)**self.plexp[k] for
            k in ('styoke_eddy', 'styoke_hyst',
                  'stteeth_eddy', 'stteeth_hyst')], axis=0)

    def iqd_plfe1(self, iq, id, f1):
        return self.betai1_plfe1(*betai1(iq, id), f1)

    def betai1_plfe2(self, beta, i1, f1):
        return np.sum([
            self._losses[k](beta, i1)*(f1/self.fo)**self.plexp[k] for
            k in ('rotor_eddy', 'rotor_hyst',)], axis=0)

    def iqd_plfe2(self, iq, id, f1):
        return self.betai1_plfe2(*betai1(iq, id), f1)

    def betai1_plmag(self, beta, i1, f1):
        return self._losses['magnet'](beta, i1)*(f1/self.fo)**2

    def iqd_plmag(self, iq, id, f1):
        return self.betai1_plmag(*betai1(iq, id), f1)


class PmRelMachinePsidq(PmRelMachine):
    """Standard set of PM machine parameters:
    p number of pole pairs
    m number of phases

    psid d-flux (Vs Peak)
    psiq q-flux (Vs Peak)
    r1 stator resistance (Ohm)
    r1 stator leakage inductance (H)
    id q current (A, Peak)
    iq q current (A, Peak)
    """

    def __init__(self, m, p, psid, psiq, r1, id, iq, ls=0, **kwargs):
        super(self.__class__, self).__init__(m, p, r1, ls)

        if isinstance(psid, (float, int)):
            self._psid = lambda id, iq: np.array([[psid]])
            self._psiq = lambda id, iq: np.array([[psiq]])
            return

        psid = np.asarray(psid)
        psiq = np.asarray(psiq)
        id = np.asarray(id)
        iq = np.asarray(iq)
        self.idrange = (min(id), max(id))
        self.iqrange = (min(iq), max(iq))
        self.betarange = (-np.pi if min(iq) < 0 else -np.pi/2,
                          0 if max(iq) > 0 else -np.pi/2)
        self.i1range = (0, np.sqrt(2)*np.min(id))
        self.io = np.max(iq)/2, np.min(id)/2

        if np.any(psid.shape < (4, 4)):
            if psid.shape[0] > 1 and psid.shape[1] > 1:
                self._psid = ip.interp2d(iq, id, psid.T)
                self._psiq = ip.interp2d(iq, id, psiq.T)
                return
            if len(id) == 1 or psid.shape[1] == 1:
                self._psid = lambda x, y: ip.InterpolatedUnivariateSpline(
                    iq, psid)(x)
                self._psiq = lambda x, y: ip.InterpolatedUnivariateSpline(
                    iq, psiq)(x)
                return
            if len(iq) == 1 or psid.shape[0] == 1:
                self._psid = lambda x, y: ip.InterpolatedUnivariateSpline(
                    id, psid)(y)
                self._psiq = lambda x, y: ip.InterpolatedUnivariateSpline(
                    id, psiq)(y)
                return
            raise ValueError("unsupported array size {}x{}".format(
                len(psid.shape[0]), psid.shape[1]))

        self._psid = lambda x, y: ip.RectBivariateSpline(
            iq, id, psid).ev(x, y)
        self._psiq = lambda x, y: ip.RectBivariateSpline(
            iq, id, psiq).ev(x, y)
        try:
            pfe = kwargs['losses']
            self._set_losspar(pfe)
            self._losses = {k: ip.RectBivariateSpline(
                iq, id, np.array(pfe[k])).ev for k in (
                'styoke_hyst', 'stteeth_hyst',
                'styoke_eddy', 'stteeth_eddy',
                'rotor_hyst', 'rotor_eddy',
                'magnet')}
        except KeyError:
            pass

    def psi(self, iq, id):
        return (self._psid(iq, id),
                self._psiq(iq, id))

    def iqdmin(self, i1):
        """max iq, min id for given current"""
        if self.idrange[0] < 0 and self.idrange[1] <= 0:
            idmin = -np.sqrt(2)*i1
        else:
            idmin = 0
        if self.idrange[0] <= idmin/np.sqrt(2):
            iqmin = -np.sqrt(2)*i1
            if self.iqrange[0] <= iqmin:
                return (iqmin, idmin)
            return self.iqrange[0], idmin

        beta = np.arccos(self.iqrange[0]/i1/np.sqrt(2))
        iqmin = np.sqrt(2)*i1*np.sin(beta)
        if self.iqrange[0] <= iqmin:
            return (iqmin, idmin)

        return self.iqrange[0], self.idrange[0]

    def iqdmax(self, i1):
        """max iq, max id for given current"""
        iqmax = np.sqrt(2)*i1
        if iqmax <= np.max(self.iqrange):
            if np.min(self.idrange) < 0 and np.max(self.idrange) <= 0:
                idmax = 0
            else:
                idmax = np.sqrt(2)*i1
            if idmax <= np.max(self.idrange):
                return (iqmax, idmax)
            return (iqmax, np.max(self.idrange))

        beta = np.arccos(self.iqrange[1]/iqmax)
        iqmax = np.sqrt(2)*i1*np.cos(beta)
        idmax = np.sqrt(2)*i1*np.sin(beta)
        if idmax <= np.max(self.idrange):
            return (iqmax, idmax)

        return iqmax, np.max(self.idrange)

    def iqd_plfe1(self, iq, id, f1):
        return np.sum([
            self._losses[k](iq, id)*(f1/self.fo)**self.plexp[k] for
            k in ('styoke_eddy', 'styoke_hyst',
                  'stteeth_eddy', 'stteeth_hyst')], axis=0)

    def betai1_plfe1(self, beta, i1, f1):
        return self.iqd_plfe1(*iqd(beta, i1), f1)

    def iqd_plfe2(self, iq, id, f1):
        return np.sum([
            self._losses[k](iq, id)*(f1/self.fo)**self.plexp[k] for
            k in ('rotor_eddy', 'rotor_hyst',)], axis=0)

    def betai1_plfe2(self, beta, i1, f1):
        return self.iqd_plfe2(*iqd(beta, i1), f1)

    def iqd_plmag(self, iq, id, f1):
        return self._losses['magnet'](iq, id)*(f1/self.fo)**2

    def betai1_plmag(self, beta, i1, f1):
        return self.iqd_plmag(*iqd(beta, i1), f1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classes for post processing based on datas from vtu-files of FEAMG-TS
"""
__version__ = '1.0.4'
__author__ = 'werner b. vetter'

import femagtools
import femagtools.nc

import vtk
import glob
import numpy as np
import netCDF4 as nc
from scipy import interpolate
import scipy.integrate as integrate
import warnings


class ts_vtu(object):
    '''Class to read vtu-files'''

    def __init__(self, modelname, directory):
        '''Read the vtu-files
        Parameters
        ----------
        modelname : str
            Name of the model (nc-file)
        directory : str
            Directory of result files (vtu-files)
        '''

        self.data = {}

        self.reader = vtk.vtkXMLUnstructuredGridReader()
        self.output = self.reader.GetOutput()

        self.field_data_names = []
        self.point_data_names = []
        self.cell_data_names = []
        self.filenames = sorted(glob.glob(
            "{0}{1}_[0-9][0-9][0-9][0-9].vtu".format(directory, modelname)))
        self.reader.SetFileName(self.filenames[0])
        self.reader.Update()
        for i in range(self.output.GetFieldData().GetNumberOfArrays()):
            self.field_data_names.append(
                self.output.GetFieldData().GetAbstractArray(i).GetName())
        for i in range(self.output.GetPointData().GetNumberOfArrays()):
            self.point_data_names.append(self.output.GetPointData().GetAbstractArray(i).GetName())
        for i in range(self.output.GetCellData().GetNumberOfArrays()):
            self.cell_data_names.append(self.output.GetCellData().GetAbstractArray(i).GetName())
            
        self.set_time_window(0.0,0.0)
        
        
    def get_data_names(self):
        '''Read the list of values stored in the vtu files
        Parameters
        ----------
            None
        Returns
        -------
            List of values stored in the vtu files
        '''
        return self.field_data_names+self.point_data_names+self.cell_data_names


    def read_data(self,data_list):
        '''Extracts data from the vtu files
        Parameters
        ----------
        data_list : fist of str
            List of values to extract from vtu_files
        '''
        for data_name in data_list:
            if data_name in self.field_data_names:
                self.data[data_name]=[]
            elif data_name in self.point_data_names:  
                self.data[data_name]=[]
            elif data_name in self.cell_data_names:  
                self.data[data_name]=[]
            else:
                raise Exception('unknown data name "'+ data_name+'"')
                
        for filename in self.filenames:            
            self.reader.SetFileName(filename)
            self.reader.Update()
            
            for data_name in data_list:
                if data_name in self.field_data_names:
                    self.data[data_name].append(self.output.GetFieldData().GetAbstractArray(data_name))
                if data_name in self.point_data_names:  
                    self.data[data_name].append(self.output.GetPointData().GetAbstractArray(data_name))
                if data_name in self.cell_data_names: 
                    self.data[data_name].append(self.output.GetCellData().GetAbstractArray(data_name))
        
        return "done"
                            

    def set_time_window(self,start,end):
        '''Set time window
        Parameters
        ----------
        start: float
            Start of the time window
        end: float
            End of the time window
            
        Only values within the time window are output by the functions
            get_field_vector
            get_point_vector
            get_cell_vector
            get_data_vector
        At start = 0.0 the values are read out starting from the first value
        At end = 0.0 the values are read out up to the last value
        '''
        try:
            if "time [s]" not in self.data:
                self.read_data(['time [s]'])
                
            if start == 0 or start <= self.data['time [s]'][0].GetValue(0):
                self.istart = 0
            else:
                self.istart = 0
                for i in range(len(self.data['time [s]'])):
                    if start >= self.data['time [s]'][i].GetValue(0):
                        self.istart = i+1 
            if end == 0 or end >= self.data['time [s]'][-1].GetValue(0):
                self.iend = len(self.data['time [s]'])
            else:
                self.iend = 0
                for i in range(len(self.data['time [s]'])):
                    if end <= self.data['time [s]'][i].GetValue(0):
                        self.iend = i
        except:
            self.istart = None
            self.iend = None

    def get_field_vector(self,field_data):
        '''Read field data
        Parameters
        ----------
        field_data : str
            Name of field to read
        Returns
        -------
        field_vec : list of float   
            List of field values within the time window
        '''
        if field_data not in self.data:
            self.read_data([field_data])

        if self.istart:
            start = self.istart
        else:
            start = 0
        if self.iend:
            end = self.iend
        else:
            end = len(self.data[field_data])
            
        field_vec=[]
        #for i in range(self.istart,self.iend):
        for i in range(start,end):
            field_vec.append(self.data[field_data][i].GetValue(0))
        return field_vec
            
    # pnt = node-key, >0
    def get_point_vector(self,pnt_data,pnt):
        '''Read point data
        Parameters
        ----------
        point_data : str
            Name of field to read
        pnt : int
            Key of point
        Returns
        -------
        point_vec : list of float   
            List of point values within the time window
        '''
        if pnt_data not in self.data:
            self.read_data([pnt_data])

        if self.istart:
            start = self.istart
        else:
            start = 0
        if self.iend:
            end = self.iend
        else:
            end = len(self.data[pnt_data])

        point_vec=[]
        for i in range(start,end):
            point_vec.append(self.data[pnt_data][i].GetValue(pnt-1))
        return point_vec
       
    def get_cell_vector(self,cell_data,cell):
        '''Read cell data
        Parameters
        ----------
        cell_data : str
            Name of field to read
        cell : int
            Key of cell
        Returns
        -------
        cell_vec : list of float   
            List of cell values within the time window
        '''
        if cell_data not in self.data:
            self.read_data([cell_data])

        i = self.cell_data_names.index(cell_data)
        noc = self.output.GetCellData().GetAbstractArray(i).GetNumberOfComponents()
        if noc ==1:
            cell_vec=[]
        else:
            cell_vec_x=[]
            cell_vec_y=[]
            cell_vec_z=[]
            
        if self.istart:
            start = self.istart
        else:
            start = 0
        if self.iend:
            end = self.iend
        else:
            end = int(len(self.data[cell_data]))

        for i in range(start,end):
            if noc == 1:
                cell_vec.append(self.data[cell_data][i].GetValue(cell-1))
            else:
                cell_vec_x.append(self.data[cell_data][i].GetValue(noc*(cell-1)))
                cell_vec_y.append(self.data[cell_data][i].GetValue(noc*(cell-1)+1))
                cell_vec_z.append(self.data[cell_data][i].GetValue(noc*(cell-1)+2))               
        if noc == 1:
            return cell_vec
        else:
            return [cell_vec_x,cell_vec_y,cell_vec_z]
        
    def get_data_vector(self,data_name,key=0):    
        '''Read data of fiels, point or cell
        Parameters
        ----------
        data_name : str
            Name of data to read
        hey : int (optional)
            Key of point or cell
        Returns
        -------
        data_vec : list of float   
            List of values within the time window
        '''
        if data_name in self.field_data_names:
            return self.get_field_vector(data_name)
        if data_name in self.point_data_names:  
            return self.get_point_vector(data_name,key)
        if data_name in self.cell_data_names:  
            return self.get_cell_vector(data_name,key)
    

class time_vector(object):
    def __init__ (self,data_sets):
        '''Read time vector in and generate an equidistant vector if necessary.
           Also the base frequency is determined.
        Parameters
        ----------
        data_sets : objects
            Data set of vtu_data, nc_data, nc_data_set
        '''
       
        try: #FEMAG-TS files
            data_list=['time [s]']
            data_sets.vtu_data.read_data(data_list)
            self.vector = data_sets.vtu_data.get_data_vector('time [s]')
            self.freq = 1/(self.vector[-1]-self.vector[0]+
                           (self.vector[1]-self.vector[0])/2+
                           (self.vector[-1]-self.vector[-2])/2)
            dt = self.vector[1]-self.vector[0]
            dt_min = 1e32
            self.equidistant = True
            for i in range(len(self.vector)-2):
                dti = self.vector[i+1]-self.vector[i]
                if dt < 0.999*dti or dt > 1.001*dti:
                    self.equidistant = False
                if dti < dt_min:
                    dt_min = dti
            if not self.equidistant:
                numpnt = int((self.vector[-1]-self.vector[0])/dt_min)
                self.vector_equi = np.linspace(self.vector[0], 
                                               self.vector[-1], 
                                               num=numpnt)
        except: #FEMAG-DC files
            speed = data_sets.nc_data_set.groups['machine'].variables['speed'][0]               
            self.freq = speed/60*data_sets.nc_data.pole_pairs
            self.equidistant = True


class losses(object):
    def __init__(self, modelname, directory):
        '''Loss calculation for FEMAG-TS simulations
        Parameters
        ----------
        modelname : str
            Name of the model (nc-file)
        directory : str
            Directory of result files (vtu-files)
        '''
        self.vtu_data = ts_vtu(modelname,directory)
        self.nc_data = femagtools.nc.read(modelname+'.nc')
        self.nc_data_set = nc.Dataset(modelname+'.nc')
        
        # self.vtu_time=object
        # self.vtu_time.vector=[]
        # self.vtu_time.equidistant = True
        
        '''Read iron losses coefficients'''
        self.iron_loss_coefficient = []
        try:
            for i in range(len(self.nc_data_set.groups['magn_curves'].variables['ch'])):
                name = self.nc_data_set.groups['mc'].variables['name'][i].tostring().decode('UTF-8')
                base_frequency = self.nc_data_set.groups['magn_curves'].variables['base_frequency'][i]
                base_induction = self.nc_data_set.groups['magn_curves'].variables['base_induction'][i]
                ch = self.nc_data_set.groups['magn_curves'].variables['ch'][i]
                ch_freq_exp = self.nc_data_set.groups['magn_curves'].variables['ch_exp'][i]
                ch_ind_exp = self.nc_data_set.groups['magn_curves'].variables['ind_exp'][i]
                kh =  ch/(base_induction**ch_ind_exp*base_frequency**ch_freq_exp)
                khml = 0.65
                cw = self.nc_data_set.groups['magn_curves'].variables['cw'][i]
                cw_freq_exp = self.nc_data_set.groups['magn_curves'].variables['cw_exp'][i]
                cw_ind_exp = self.nc_data_set.groups['magn_curves'].variables['ind_exp'][i]
                if cw_freq_exp == cw_ind_exp:
                    kw = self.coeff_frequency_to_time(base_induction,base_frequency,cw,cw_freq_exp)
                else:
                    warnings.warn('Waterfall methode not possible, specify parameter kw')
                    kw = 0.0
                ce = 0.0
                ce_freq_exp = 1.5
                ce_ind_exp = 1.5
                ke = 0.0
                spec_weight = self.nc_data_set.groups['magn_curves'].variables['spec_weight'][i]
                fillfactor = self.nc_data_set.groups['magn_curves'].variables['fillfac'][i]
                shapefactor = self.nc_data_set.groups['loss_coeffs'].variables['shape_factor'][i]

                coeffdict = {
                    "Name": np.char.strip(name).tolist(),
                    "base_frequency": base_frequency.tolist(),
                    "base_induction": base_induction.tolist(),
                    "ch": ch.tolist(),
                    "ch_freq_exp": ch_freq_exp.tolist(),
                    "ch_ind_exp": ch_ind_exp.tolist(),
                    "kh": kh,
                    "khml":khml,
                    "cw": cw.tolist(),
                    "cw_freq_exp": cw_freq_exp.tolist(),
                    "cw_ind_exp": cw_ind_exp.tolist(),
                    "kw": kw,
                    "ce": ce,
                    "ce_freq_exp": ce_freq_exp,
                    "ce_ind_exp": ce_ind_exp,
                    "ke": ke,
                    "spec_weight": spec_weight.tolist(),
                    "fillfactor": fillfactor.tolist(),
                    "shapefactor": shapefactor.tolist()
                }
                self.iron_loss_coefficient.append(coeffdict)
        except:
            pass
            
        '''Read loss coefficients'''    
        for i in range(2):
            if i == 0:
                name = 'Outside'
            else:
                name = 'Inside'
                
            base_frequency = self.nc_data_set.groups['loss_coeffs'].variables['freq_0'][i]
            base_induction = self.nc_data_set.groups['loss_coeffs'].variables['ind_0'][i]
            ch = self.nc_data_set.groups['loss_coeffs'].variables['ch'][i]
            ch_freq_exp = self.nc_data_set.groups['loss_coeffs'].variables['fa_hyst'][i]
            ch_ind_exp = self.nc_data_set.groups['loss_coeffs'].variables['fa_ind'][i]
            kh =  ch/(base_induction**ch_ind_exp*base_frequency**ch_freq_exp)
            khml = 0.65
            cw = self.nc_data_set.groups['loss_coeffs'].variables['cw'][i]
            cw_freq_exp = self.nc_data_set.groups['loss_coeffs'].variables['fa_eddy'][i]
            cw_ind_exp = self.nc_data_set.groups['loss_coeffs'].variables['fa_ind'][i]
            if cw_freq_exp == cw_ind_exp:
                kw = self.coeff_frequency_to_time(base_induction,base_frequency,cw,cw_freq_exp)
            else:
                warnings.warn('Waterfall methode not possible, specify parameter kw')
                kw = 0.0
            ce = 0.0
            ce_freq_exp = 1.5
            ce_ind_exp = 1.5
            ke = 0.0
            spec_weight = self.nc_data_set.groups['loss_coeffs'].variables['spec_mass_fe'][i]
            fillfactor = self.nc_data_set.groups['loss_coeffs'].variables['fill_factor_fe'][i]
            shapefactor = self.nc_data_set.groups['loss_coeffs'].variables['shape_factor'][i]

            coeffdict = {
                "Name": name,
                "base_frequency": base_frequency.tolist(),
                "base_induction": base_induction.tolist(),
                "ch": ch.tolist(),
                "ch_freq_exp": ch_freq_exp.tolist(),
                "ch_ind_exp": ch_ind_exp.tolist(),
                "kh":kh,
                "khml":khml,
                "cw": cw.tolist(),
                "cw_freq_exp": cw_freq_exp.tolist(),
                "cw_ind_exp": cw_ind_exp.tolist(),
                "kw": kw,
                "ce": ce,
                "ce_freq_exp": ce_freq_exp,
                "ce_ind_exp": ce_ind_exp,
                "ke": ke,
                "spec_weight": spec_weight.tolist(),
                "fillfactor": fillfactor.tolist(),
                "shapefactor": shapefactor.tolist()
            }
            self.iron_loss_coefficient.append(coeffdict)
            

    def scale_factor(self):
        '''Returns the scale dactor
        Parameters
        ----------
        None
            
        Returns
        -------
        scale_facor : int
        '''
        try:
            poles = 2*self.nc_data.pole_pairs
        except:
            poles = 2
        try:
            poles_sim = self.nc_data.poles_sim
        except:
            poles_sim = poles
            
        scale_factor = poles/poles_sim
            
        return scale_factor
    
    def coeff_frequency_to_time(self,B0,f0,c,exp):
        '''Convert Bertotti-coefficient of frequency domain to time domains coefficient
        Parameters
        ----------
        B0 : float
            Base flux density [T]
        f0 : float
            Base freuency [Hz]
        c : float
            Bertotti-coefficient
        exp : float
            Bertotti-exponent
        Return
        -------
        k : float   
            Losse coefficient in time domains
            
        The conversion is only possible for losses-coefficient with
        equal exponent for flux density and frequency,
        as eddy current losses (cw*(B/B0(*)*2*(f/f0)**2) or
        anomalous losses (ce*(B/B0(*)**1.5*(f/f0)**1.5)
        '''
        factor = (2*np.pi)**exp*integrate.quad(lambda x: np.abs(np.cos(2*np.pi*x))**(exp), 0, 1)[0]
        k = c/(B0**exp*f0**exp)/factor
        return k
            
              
#ohmic losses
    def ohm_lossenergy_sr(self,sr):
        '''Ohmic loss energy of a subregion
        Parameters
        ----------
        sr : object
            Subregion
            
        Returns
        -------
        lossenergy : float
            Ohmic loss energy of the subregion
        
        The loss energy is determined by adding up the loss energy of the
        individual elements.
        '''
        scale_factor = self.scale_factor()
        length = self.nc_data.arm_length
        time = self.time_vector
        
        srlossenergy = 0.0
        for supel in sr.superelements:
            selossenergy = 0.0
            if supel.conduc > 0.0:
                ff=self.nc_data_set.groups['superelements'].variables['fillfactor'][supel.key-1]
                if ff == 0.0:
                    ff = 1.0
                #print(supel.key, supel.conduc, supel.length, ff)
                for el in supel.elements:
                    #print(el.key,  el.area)  
                    ellossenergy = 0.0
                    cd_vec=self.vtu_data.get_data_vector('curd',el.key)
                    for j in range(len(time)-1):
                        cd = (cd_vec[j]+cd_vec[j+1])/2
                        dt = time[j+1]-time[j]
                        ellossenergy = ellossenergy + dt*cd**2*el.area/ff/supel.conduc*supel.length
                    selossenergy = selossenergy + ellossenergy*length*scale_factor         
            
            srlossenergy = srlossenergy + selossenergy
                
        return srlossenergy


    def ohm_lossenergy_subregion(self,srname,start=0.0,end=0.0):
        '''Ohmic loss energy of a subregion
        Parameters
        ----------
        srname:  str
            Name of subregion
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
            
        Returns
        -------
        lossenergy : float
            Ohmic loss energy of the subregion
        
        The loss energy is determined by adding up the loss energy of the
        individual elements over the time window.
        If start and end are not specified, the time window of the 
        previous calculation is used.
        '''       
        while len(srname)<4:
            srname = srname+' '
            
        data_list=['time [s]','curd']
        self.vtu_data.read_data(data_list)

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start,end)
            
        self.time_vector = self.vtu_data.get_data_vector('time [s]')
        
        sr = self.nc_data.get_subregion(srname)
        srlossenergy = self.ohm_lossenergy_sr(sr)
        return srlossenergy
    
    def ohm_powerlosses_subregion(self,srname,start=0.0,end=0.0):
        '''Ohmic loss dissipation of a subregion within the time window
        Parameters
        ----------
        srname : str
            Name of subregion
        start : float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
            
        Returns
        -------
        powerlosses : float
            Ohmic loss dissipation of the subregion
        
        The loss energy is determined by adding up the loss energy of the
        individual elements over the time window.
        The loss energy is divided by the time window length
        to obtain the averaged power loss
        If start and end are not specified, the time window of the 
        previous calculation is used.
        '''
        while len(srname)<4:
            srname = srname+' '
            
        srlossenergy = self.ohm_lossenergy_subregion(srname,start,end)       
        srpowerlosses = srlossenergy/(self.time_vector[-1]-self.time_vector[0])
        return srpowerlosses
        
    def ohm_lossenergy(self,start=0.0,end=0.0):
        '''Ohmic loss energy of all subregions
        Parameters
        ----------
        start: float
            Start of the time window (optional)
        end: float
            End of the time window (optional)
            
        Returns
        -------
        loss_data: dict
            Dictonary of subregions and ohmic loss energy of it
        
        The loss energy is determined by adding up the loss energy of the
        individual elements over the time window.
        If start and end are not specified, the time window of the 
        previous calculation is used.
        '''
        
        data_list=['time [s]','curd']
        self.vtu_data.read_data(data_list)

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start,end)
            
        self.time_vector = self.vtu_data.get_data_vector('time [s]')
       
        loss_data = []
        for sr in self.nc_data.subregions:
            srlossenergy = self.ohm_lossenergy_sr(sr)
            
            srname = sr.name
            if sr.wb_key >= 0:
                #print(sr.key,"is winding",sr.wb_key+1)
                if srname == '    ':
                    srname = "wdg "+str(sr.wb_key+1)
                        
            loss_data.append({'key':sr.key, 'name':srname, 'losses':srlossenergy})

        return loss_data

    def ohm_powerlosses(self,start=0.0,end=0.0):
        '''Ohmic loss dissipation of all subregions
        Parameters
        ----------
        start : float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
            
        Returns
        -------
        loss_data : dict
            Dictonary of subregions and ohmic loss dissipation of it
        
        The loss energy is determined by adding up the loss energy of the
        individual elements over the time window.
        The loss energy is divided by the time window length
        to obtain the averaged power loss
        If start and end are not specified, the time window of the 
        previous calculation is used.
        '''

        data_list=['time [s]','curd']
        self.vtu_data.read_data(data_list)

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start,end)
        
        self.time_vector = self.vtu_data.get_data_vector('time [s]')
         
        loss_data = []
        for sr in self.nc_data.subregions:
            srlossenergy = self.ohm_lossenergy_sr(sr)
            srpowerlosses = srlossenergy/(self.time_vector[-1]-self.time_vector[0])
            
            srname = sr.name
            if sr.wb_key >= 0:
                #print(sr.key,"is winding",sr.wb_key+1)
                if srname == '    ':
                    srname = "wdg "+str(sr.wb_key+1)
                        
            loss_data.append({'key':sr.key, 'name':srname, 'losses':srpowerlosses})

        return loss_data

    def ohm_powerlosses_fft_sr(self,sr):
        '''Power dissipation of a subregion
        Parameters
        ----------
        sr : object
            Subregion
            
        Returns
        -------
        powerlosses : float
            Ohmic power losses of the subregion
        
        A FFT from the current density is made. 
        The power losses of each harmonic is determined and added.
        '''
        scale_factor = self.scale_factor()
        length = self.nc_data.arm_length
                
        srpowerlosses = 0.0
        for supel in sr.superelements:
            sepowerlosses = 0.0
            if supel.conduc > 0.0:
                ff=self.nc_data_set.groups['superelements'].variables['fillfactor'][supel.key-1]
                if ff == 0.0:
                    ff = 1.0
                    #print(supel.key, supel.conduc, supel.length, ff)
                for el in supel.elements:
                    #print(el.key,  el.area)  
                    elpowerlosses = 0.0
                    cd_vec_0 = self.vtu_data.get_data_vector('curd',el.key)
                    if not self.times.equidistant:
                        cd_vec = np.interp(self.times.vector_equi, 
                                            self.times.vector, cd_vec_0, 
                                            period=1.0/self.times.freq)
                        # f = interpolate.interp1d(self.times.vector, cd_vec_0, kind="cubic")
                        # cd_vec = f(self.times.vector_equi)
                    else:
                        cd_vec = cd_vec_0
                    cd_spec = abs(np.fft.fft(cd_vec))/(len(cd_vec)/2)               
                    for j in range(int(len(cd_vec)/2)):
                        elpowerlosses = elpowerlosses + cd_spec[j]**2/2*el.area/ff/supel.conduc*supel.length

                    sepowerlosses = sepowerlosses + elpowerlosses*length*scale_factor
                    
            srpowerlosses = srpowerlosses + sepowerlosses

        return srpowerlosses

    def ohm_powerlosses_fft_subregion(self,srname,start=0.0,end=0.0):
        '''Power dissipation of a subregion
        Parameters
        ----------
        srname:  str
            Name of subregion
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
            
        Returns
        -------
        lossenergy : float
            Power dissipation of the subregion
        
        A FFT from the current density is made. 
        The power losses of each harmonic is determined and added.
        The time window has to be pariode or a multiple of it.
        If start and end are not specified, the time window of the 
        previous calculation is used.
        '''
        while len(srname)<4:
            srname = srname+' '
                    
        data_list=['time [s]','curd']
        self.vtu_data.read_data(data_list)

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start,end)
            
        self.times = time_vector(self)
        
        sr = self.nc_data.get_subregion(srname)
        srpowerlosses = self.ohm_powerlosses_fft_sr(sr)
        return srpowerlosses


    def ohm_powerlosses_fft(self,start=0.0,end=0.0):
        '''Power dissipation of all subregions
        Parameters
        ----------
        start : float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
           
        Returns
        -------
        loss_data : dict
            Dictonary of subregions and power dissipation of it
       
        A FFT from the current density is made. 
        The power losses of each harmonic is determined and added.
        The time window has to be pariode or a multiple of it.
        If start and end are not specified, the time window of the 
        previous calculation is used.
        '''
        
        data_list=['time [s]','curd']
        self.vtu_data.read_data(data_list)

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start,end)
       
        self.times = time_vector(self)

        loss_data = []
        for sr in self.nc_data.subregions:
            srpowerlosses= self.ohm_powerlosses_fft_sr(sr)
            
            srname = sr.name
            if sr.wb_key >= 0:
                #print(sr.key,"is winding",sr.wb_key+1)
                if srname == '    ':
                    srname = "wdg "+str(sr.wb_key+1)
                        
            loss_data.append({'key':sr.key, 'name':srname, 'losses':srpowerlosses})

        return loss_data

# iron losses 
    def iron_losses_fft_se(self,se):
        '''Iron losses of a superelement
        Parameters
        ----------
        se: object
            Superelement
            
        Returns
        -------
        ironlosses : float
            Iron losses of the superlement
        
        A FFT is made from the flux density. 
        The iron losses of each harmonic is determined  by 
        Bertotti formula 
        
            Physt = ch * (f/f0)**hfe * (B/B0)**hBe * V * rho
            Peddy = ch * (f/f0)**wfe * (B/B0)**wBe * V * rho
            Pexce = ch * (f/f0)**efe * (B/B0)**eBe * V * rho
            
        and added to the total losses of the superelement
        
            Ptot  = (Physt + Peddy + Pexce) * shape_factor
        '''

        scale_factor = self.scale_factor()
        length = self.nc_data.arm_length
        freq = self.times.freq
        
        sehystlosses = 0.0
        seeddylosses = 0.0
        seexcelosses = 0.0
        if se.elements[0].reluc[0]<1.0 or se.elements[0].reluc[1]<1.0:
            if (se.mcvtype == 0):
                center_pnt = se.elements[0].center
                if (np.sqrt(center_pnt[0]**2+center_pnt[1]**2) > self.nc_data.FC_RADIUS):
                    ldi = len(self.iron_loss_coefficient)-2  #outside
                else:
                    ldi = len(self.iron_loss_coefficient)-1  #inside
            else:
                ldi = se.mcvtype-1
            bf = self.iron_loss_coefficient[ldi]['base_frequency']
            bb = self.iron_loss_coefficient[ldi]['base_induction']
            ch = self.iron_loss_coefficient[ldi]['ch']
            chfe = self.iron_loss_coefficient[ldi]['ch_freq_exp']
            chbe = self.iron_loss_coefficient[ldi]['ch_ind_exp']
            cw = self.iron_loss_coefficient[ldi]['cw']
            cwfe= self.iron_loss_coefficient[ldi]['cw_freq_exp']
            cwbe= self.iron_loss_coefficient[ldi]['cw_ind_exp']
            ce = self.iron_loss_coefficient[ldi]['ce']
            cefe = self.iron_loss_coefficient[ldi]['ce_freq_exp']
            cebe = self.iron_loss_coefficient[ldi]['ce_ind_exp']
            sw = self.iron_loss_coefficient[ldi]['spec_weight']*1000
            ff = self.iron_loss_coefficient[ldi]['fillfactor']
            sf = self.iron_loss_coefficient[ldi]['shapefactor']
 
            for el in se.elements:
                #print(el.key,  el.area)             
                elhystlosses = 0.0
                eleddylosses = 0.0
                elexcelosses = 0.0
                
                bx_vec_0 = self.vtu_data.get_data_vector('b',el.key)[0]
                if not self.times.equidistant:
                    bx_vec = np.interp(self.times.vector_equi, 
                                        self.times.vector, bx_vec_0, 
                                        period=1.0/self.times.freq)
                    # f = interpolate.interp1d(self.times.vector, bx_vec_0, kind="cubic")
                    # bx_vec = f(self.times.vector_equi)
                else:
                    bx_vec = bx_vec_0
                bx_spec = abs(np.fft.fft(bx_vec))/(len(bx_vec)/2)    
                
                by_vec_0 = self.vtu_data.get_data_vector('b',el.key)[1]
                if not self.times.equidistant:
                    by_vec = np.interp(self.times.vector_equi, 
                                        self.times.vector, by_vec_0, 
                                        period=1.0/self.times.freq)
                    # f = interpolate.interp1d(self.times.vector, by_vec_0, kind="cubic")
                    # by_vec = f(self.times.vector_equi)
                else:
                    by_vec = by_vec_0
                by_spec = abs(np.fft.fft(by_vec))/(len(by_vec)/2)  
                
                b_spec = np.sqrt((bx_spec**2+by_spec**2))

                for j in range(int(len(b_spec)/2)):
                    elhystlosses = elhystlosses + ch*(j*freq/bf)**chfe*(b_spec[j]/bb)**chbe
                    eleddylosses = eleddylosses + cw*(j*freq/bf)**cwfe*(b_spec[j]/bb)**cwbe
                    elexcelosses = elexcelosses + ce*(j*freq/bf)**cefe*(b_spec[j]/bb)**cebe
                   
                sehystlosses = sehystlosses + elhystlosses*el.area*length*ff*sf*sw*scale_factor
                seeddylosses = seeddylosses + eleddylosses*el.area*length*ff*sf*sw*scale_factor
                seexcelosses = seexcelosses + elexcelosses*el.area*length*ff*sf*sw*scale_factor
                 
        setotallosses = sehystlosses + seeddylosses + seexcelosses
        
        return {'total':setotallosses,
                'hysteresis':sehystlosses,
                'eddycurrent':seeddylosses,
                'excess':seexcelosses}


    def iron_losses_fft_subregion(self,srname,start=0.0,end=0.0):
        '''Iron losses of a subregion
        Parameters
        ----------
        srname:  str
            Name of subregion
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
            
        Returns
        -------
        losses : dict
            Iron losses of the subregion
        
        The iron losses are calculated based on the Bertotti formula
        (see also ron_losses_fft_se)
        '''
        while len(srname)<4:
            srname = srname+' '
            
        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start,end)
        
        data_list=['b']
        self.vtu_data.read_data(data_list)
        self.times = time_vector(self)

        srtotallosses = 0.0
        srhystlosses = 0.0
        sreddylosses = 0.0
        srexcelosses = 0.0
        sr = self.nc_data.get_subregion(srname)
        for se in sr.superelements:
            selosses = self.iron_losses_fft_se(se)
            srtotallosses = srtotallosses + selosses['total']
            srhystlosses = srhystlosses + selosses['hysteresis']
            sreddylosses = sreddylosses + selosses['eddycurrent']
            srexcelosses = srexcelosses + selosses['excess']
            
        srlosses = {'subregion':srname,
                    'total':srtotallosses,
                    'hysteresis':srhystlosses,
                    'eddycurrent':sreddylosses,
                    'excess':srexcelosses 
                    }
        return srlosses

    def iron_losses_fft(self,start=0.0,end=0.0):
        '''Iron losses of all subregion and superelements
        Parameters
        ----------
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
            
        Returns
        -------
        losses : dict
            Iron losses of the subregion
        
        The iron losses are calculated based on the Bertotti formula
        (see also iron_losses_fft_se)
        '''
        
        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start,end)
        
        data_list=['b']
        self.vtu_data.read_data(data_list)        
        self.times = time_vector(self)
            
        losseslist = []
 
        for se in self.nc_data.superelements:
            selosses = self.iron_losses_fft_se(se)

            if se.subregion:
                for sr in self.nc_data.subregions:
                    if se in sr.superelements:
                        srname = sr.name
                        #print(se.key, "in", sr.key, sr.name)                
            else:
                if (se.mcvtype == 0):
                    center_pnt = se.elements[0].center
                    if (np.sqrt(center_pnt[0]**2+center_pnt[1]**2) > self.nc_data.FC_RADIUS):
                        srname = "no, outside"
                    else:
                        srname = "no, inside"
                        
            found = False
            for srlosses in losseslist:
                if srlosses['subregion'] == srname:
                        srlosses['total'] = srlosses['total']+selosses['total']
                        srlosses['hysteresis'] = srlosses['hysteresis']+selosses['hysteresis']
                        srlosses['eddycurrent'] = srlosses['eddycurrent']+selosses['eddycurrent']
                        srlosses['excess'] = srlosses['excess']+selosses['excess']
                        found = True
            if not found:
                if selosses['total'] > 0.0:
                    srlosses = {'subregion':srname,
                        'total':selosses['total'],
                        'hysteresis':selosses['hysteresis'],
                        'eddycurrent':selosses['eddycurrent'],
                        'excess':selosses['excess']
                        }
                    losseslist.append(srlosses)                         

        return losseslist


    def iron_lossenergy_time_se(self,se):
        '''Iron losses of a superelement in time domain
        Parameters
        ----------
        se: object
            Superelement
            
        Returns
        -------
        lossenergies : float
            Iron losses of the superlement
        
        The iron losses are calculated based on the Bertotti formula
        in time domaine.
        The loss coefficients in frequency domain are converted into 
        time domain coefficients.
        For the hysteresis losses is a water fall methode implemented.
        Eddy current losses and anomalous losses are calculated by
        add up the losses of each time step.
        '''

        scale_factor = self.scale_factor()
        length = self.nc_data.arm_length
        time = self.times.vector
        
        sehystenergy = 0.0
        seeddyenergy = 0.0
        seexceenergy = 0.0
        if se.elements[0].reluc[0]<1.0 or se.elements[0].reluc[1]<1.0:
            if (se.mcvtype == 0):
                center_pnt = se.elements[0].center
                if (np.sqrt(center_pnt[0]**2+center_pnt[1]**2) > self.nc_data.FC_RADIUS):
                    ldi = len(self.iron_loss_coefficient)-2  #outside
                else:
                    ldi = len(self.iron_loss_coefficient)-1  #inside
            else:
                ldi = se.mcvtype-1
            kh = self.iron_loss_coefficient[ldi]['kh']
            chbe = self.iron_loss_coefficient[ldi]['ch_ind_exp']
            khml = self.iron_loss_coefficient[ldi]['khml']
            kw = self.iron_loss_coefficient[ldi]['kw']
            cwbe = self.iron_loss_coefficient[ldi]['cw_ind_exp']
            ke = self.iron_loss_coefficient[ldi]['ke']
            cebe = self.iron_loss_coefficient[ldi]['ce_ind_exp']
            sw = self.iron_loss_coefficient[ldi]['spec_weight']*1000
            ff = self.iron_loss_coefficient[ldi]['fillfactor']
            sf = self.iron_loss_coefficient[ldi]['shapefactor']
 
            for el in se.elements:
                elhystenergy = 0.0
                eleddyenergy = 0.0
                elexceenergy = 0.0
                
                bx_vec = self.vtu_data.get_data_vector('b',el.key)[0]
                by_vec = self.vtu_data.get_data_vector('b',el.key)[1]                

                #Maximalwert und Richtung des Haupfeldes
                Bpeak = np.sqrt(bx_vec[0]**2+by_vec[0]**2)
                phi = np.arctan2(by_vec[0],bx_vec[0])
                for i in range(1,len(time)):
                    b1 = np.sqrt(bx_vec[i-1]**2+by_vec[i-1]**2)
                    b2 = np.sqrt(bx_vec[i]**2+by_vec[i]**2)
                    if abs(b2)>Bpeak:
                        Bpeak = abs(b2)
                        phi = np.arctan2(by_vec[i],bx_vec[i])
                        
                #Transformation in Hauptrichutng
                br_vec = []
                bt_vec = []
                for i in range(len(time)):   
                    br_vec.append(np.cos(phi)*bx_vec[i]+np.sin(phi)*by_vec[i])
                    bt_vec.append(np.sin(phi)*bx_vec[i]-np.cos(phi)*by_vec[i])
                        
                Bpeak_p = np.sqrt(bx_vec[0]**2+by_vec[0]**2)
                Bx = []
                tp_beg = 0.0
                tp_end = 0.0
                Tp = 0.0
                nzeros = 0
                zero = (br_vec[0]>=0)
                if br_vec[1]>br_vec[0]:
                    up = True
                else:
                    up = False
                for i in range(1,len(time)):
                    b1 = np.sqrt(br_vec[i-1]**2+bt_vec[i-1]**2)
                    b2 = np.sqrt(br_vec[i]**2+bt_vec[i]**2)
                    #Maximalwert innerhalb letzter Periode
                    if abs(b2)>Bpeak_p:
                        Bpeak_p = abs(b2)
                    #Nulldurchgaenge und Periodendauer
                    if zero != (br_vec[i]>=0):
                        zero = (not zero)
                        tp_beg = tp_end
                        tp_end = time[i]
                        if tp_beg > 0.0:
                            nzeros = nzeros+1
                            if nzeros > 1:
                                #Tp = (Tp*(nzeros-1)/nzeros+2*(tp_end-tp_beg)/nzeros)/2
                                Tp = 2*(tp_end-tp_beg)
                                Bpeak = Bpeak_p
                                elhystenergy = elhystenergy+kh*Bpeak**chbe/2
                                Bpeak_p = 0.0
                            else:
                                Tp = 2.0*(tp_end-tp_beg)
                                Bpeak = Bpeak_p
                                elhystenergy = elhystenergy+kh*Bpeak**chbe * (tp_end-time[0])/Tp
                                Bpeak_p = 0.0
                        Bx = []
                    #Wendepunkte       
                    if up and b2<b1:
                        Bx.append(b1)
                    if not up and b2>b1:
                        Bx.append(b1)
                    #Steigungsrichtung       
                    if b2>b1:
                        up = True
                    else:
                        up = False
             
                    try:
                        if b2>0 and up and b2>Bx[-2]:
                            Bm = abs(Bx[-2]+Bx[-1])/2
                            dB = abs(Bx[-2]-Bx[-1])
                            elhystenergy = elhystenergy+kh*Bm**(chbe-1)*khml*dB/2
                            Bx.remove(Bx[-2])
                            Bx.remove(Bx[-1])
                        if b2<0 and not up and b2<Bx[-2]:
                            elhystenergy = elhystenergy+kh*Bm**(chbe-1)*khml*dB/2
                            Bx.remove(Bx[-2])
                            Bx.remove(Bx[-1])
                        if b2>0 and not up and Bx[-1]>Bx[-2]:
                            elhystenergy = elhystenergy+kh*Bm**(chbe-1)*khml*dB/2
                            Bx.remove(Bx[-2])
                            Bx.remove(Bx[-1])
                        if b2<0 and up and Bx[-1]<Bx[-2]:                           
                            elhystenergy = elhystenergy+kh*Bm**(chbe-1)*khml*dB/2
                            Bx.remove(Bx[-2])
                            Bx.remove(Bx[-1])
                                  
                    except:
                        pass
                    
                    dt = time[i]-time[i-1]
                    dbr = br_vec[i]-br_vec[i-1]
                    dbt = bt_vec[i]-bt_vec[i-1]
                    db = np.sqrt(dbr**2+dbt**2)
                    eleddyenergy = eleddyenergy + kw*(db/dt)**cwbe * dt
                    elexceenergy = elexceenergy + ke*(db/dt)**cebe * dt
                    
                #elhystenergy = elhystenergy+kh*Bpeak**chbe * T/(time[-1]-time[0])
                if nzeros >= 1:
                    elhystenergy = elhystenergy+kh*Bpeak**chbe * (time[-1]-tp_end)/Tp
                
                sehystenergy = sehystenergy + elhystenergy*el.area*length*ff*sf*sw*scale_factor
                seeddyenergy = seeddyenergy + eleddyenergy*el.area*length*ff*sf*sw*scale_factor
                seexceenergy = seexceenergy + elexceenergy*el.area*length*ff*sf*sw*scale_factor
                 
        setotalenergy = sehystenergy + seeddyenergy + seexceenergy
        
        return {'total':setotalenergy,
                'hysteresis':sehystenergy,
                'eddycurrent':seeddyenergy,
                'excess':seexceenergy}

    def iron_lossenergy_time_subregion(self,srname,start=0.0,end=0.0):
        '''Iron loss energy of a subregion
        Parameters
        ----------
        srname:  str
            Name of subregion
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
            
        Returns
        -------
        losses : dict
            Iron losses energy of the subregion
        
        The iron losses are calculated based on the Bertotti formula
        in time domaine (see also iron_lossenergy_time_se)
        '''
        while len(srname)<4:
            srname = srname+' '
            
        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start,end)
        
        data_list=['b']
        self.vtu_data.read_data(data_list)
        self.times = time_vector(self)

        srtotalenergy = 0.0
        srhystenergy = 0.0
        sreddyenergy = 0.0
        srexceenergy = 0.0
        sr = self.nc_data.get_subregion(srname)
        for se in sr.superelements:
            seenergy = self.iron_lossenergy_time_se(se)
            srtotalenergy = srtotalenergy + seenergy['total']
            srhystenergy = srhystenergy + seenergy['hysteresis']
            sreddyenergy = sreddyenergy + seenergy['eddycurrent']
            srexceenergy = srexceenergy + seenergy['excess']
            
        srenergy = {'subregion':srname,
                    'total':srtotalenergy,
                    'hysteresis':srhystenergy,
                    'eddycurrent':sreddyenergy,
                    'excess':srexceenergy 
                    }
        return srenergy

    def iron_losses_time_subregion(self,srname,start=0.0,end=0.0):
        '''Iron power losses of a subregion
        Parameters
        ----------
        srname:  str
            Name of subregion
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
            
        Returns
        -------
        losses : dict
            Iron losses energy of the subregion
        
        The iron losses are calculated based on the Bertotti formula
        in time domaine (see also iron_lossenergy_time_se)
        '''
        while len(srname)<4:
            srname = srname+' '

        srenergy = self.iron_lossenergy_time_subregion(srname,start,end)
        time = self.times.vector[-1]-self.times.vector[0]
        
        srlosses = {'subregion':srname,
                    'total':srenergy['total']/time,
                    'hysteresis':srenergy['hysteresis']/time,
                    'eddycurrent':srenergy['eddycurrent']/time,
                    'excess':srenergy['excess']/time
                    }
                                          
        return srlosses
        
    def iron_lossenergy_time(self,start=0.0,end=0.0):
        '''Iron losses of all subregion and superelements
        Parameters
        ----------
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
            
        Returns
        -------
        energies : dict
            Iron losses enegies of the subregion
        
        The iron losses are calculated based on the Bertotti formula
        in time domaine (see also iron_lossenergy_time_se)
        '''
        
        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start,end)
        
        data_list=['b']
        self.vtu_data.read_data(data_list)        
        self.times = time_vector(self)
            
        energylist = []
 
        for se in self.nc_data.superelements:
            selossenergy = self.iron_lossenergy_time_se(se)

            if se.subregion:
                for sr in self.nc_data.subregions:
                    if se in sr.superelements:
                        srname = sr.name
                        #print(se.key, "in", sr.key, sr.name)                
            else:
                if (se.mcvtype == 0):
                    center_pnt = se.elements[0].center
                    if (np.sqrt(center_pnt[0]**2+center_pnt[1]**2) > self.nc_data.FC_RADIUS):
                        srname = "no, outside"
                    else:
                        srname = "no, inside"
                        
            found = False
            for srlosses in energylist:
                if srlosses['subregion'] == srname:
                        srlosses['total'] = srlosses['total']+selossenergy['total']
                        srlosses['hysteresis'] = srlosses['hysteresis']+selossenergy['hysteresis']
                        srlosses['eddycurrent'] = srlosses['eddycurrent']+selossenergy['eddycurrent']
                        srlosses['excess'] = srlosses['excess']+selossenergy['excess']
                        found = True
            if not found:
                if selossenergy['total'] > 0.0:
                    srlosses = {'subregion':srname,
                        'total':selossenergy['total'],
                        'hysteresis':selossenergy['hysteresis'],
                        'eddycurrent':selossenergy['eddycurrent'],
                        'excess':selossenergy['excess']
                        }
                    energylist.append(srlosses)                         

        return energylist

    def iron_losses_time(self,start=0.0,end=0.0):
        '''Iron losses of all subregion and superelements
        Parameters
        ----------
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)
            
        Returns
        -------
        losses : dict
            Iron losses of the subregion
        
        The iron losses are calculated based on the Bertotti formula
        in time domaine (see also iron_lossenergy_time_se)
        '''
        
        energylist = self.iron_lossenergy_time(start,end)
        time = self.times.vector[-1]-self.times.vector[0]
        
        losseslist = []
        for sr in energylist:
            sr['total'] = sr['total']/time
            sr['hysteresis'] = sr['hysteresis']/time
            sr['eddycurrent'] = sr['eddycurrent']/time
            sr['excess'] = sr['excess']/time
            losseslist.append(sr)
            
        return losseslist



# -*- coding: utf-8 -*-
"""
    femagtools.model
    ~~~~~~~~~~~~~~~~

    Managing model parameters



"""
import logging
import string
import numpy as np

logger = logging.getLogger(__name__)
#
# Set of legal model name chars
# 0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ
MODELNAME_CHARS = string.printable[:63] + "-_äöüéè"
# maximum name length
MAX_MODEL_NAME_LEN = 255


def movesteps(nodes):
    """returns list of move steps

    Args:
       nodes: number of nodes in airgap
    """
    return [nodes // w + 1 for w in range(1, nodes//2) if nodes % w == 0]


class MCerror(Exception):
    pass


class Model(object):
    def __init__(self, parameters):
        if isinstance(parameters, dict):
            for k in parameters.keys():
                setattr(self, k, parameters[k])

    def set_value(self, name, value, p=None):
        """set value of parameter identified by name

        Args:
            name: name of parameter
            value: value to be assigned to parameter
        """
        if isinstance(name, str):
            setattr(self, name, value)
            return

        if len(name) > 1:
            k = name[0]
            if hasattr(self, k):
                self.set_value(name[1:], value, getattr(self, k))
                return
            elif p:
                self.set_value(name[1:], value, p[k])
                return
            self.set_value(name[1:], value, self)
            return
        if p:
            p[name[0]] = value
            return
        setattr(self, name[0], value)

    def get(self, name, r=None):
        """return value of key name

        Args:
            name: key of parameter value

        Return:
            value of parameter identified by key
        """
        try:
            if isinstance(name, str):
                if r is not None:
                    return getattr(self, name, r)
                return getattr(self, name)
            if r and type(r) == dict:
                for k in name:
                    r = r.get(k)
                return r
            if len(name) > 1:
                if r:
                    return self.get(name[1:], getattr(r, name[0]))
                return self.get(name[1:], getattr(self, name[0]))
            return getattr(self, name[0])
        except KeyError as e:
            logger.error(e)
            raise MCerror(e)

    def __getitem__(self, name):
        return getattr(self, name)

    def __str__(self):
        "return string format of this object"
        return repr(self.__dict__)

    def __repr__(self):
        "representation of this object"
        return self.__str__()


class MachineModel(Model):
    """represents a machine model for a FE analysis

    Args:
      parameters: string or dict containing the model parameters. For example:
    ::

        {'lfe': 0.1,
        'poles': 4,
        'outer_diam': 0.13,
        'bore_diam': 0.07,
        'stator':{
        'num_slots': 12,
        'num_slots_gen': 3,
        ...
        },
        'magnet':{
            'material': 'M395',
            ..
        }
        }

        if parameters is string it is interpreted as the model name.
    """

    def __init__(self, parameters):
        super(self.__class__, self).__init__(parameters)
        name = 'DRAFT'
        if isinstance(parameters, str):
            name = parameters
        elif 'name' in parameters:
            name = parameters['name']
        # connect model even for complete model (see fsl connect_models)
        self.connect_full = True
        # must sanitize name to prevent femag complaints
        self.name = ''.join([n
                             for n in name.strip()
                             if n in MODELNAME_CHARS][:MAX_MODEL_NAME_LEN])
        try:
            self.external_rotor = (self.external_rotor == 1)
        except AttributeError:
            self.external_rotor = False
        self.move_inside = 1.0 if self.external_rotor else 0.0
        if 'exit_on_end' not in parameters:
            self.exit_on_end = 'false'
        if 'magnet' in parameters:
            if 'mcvkey_mshaft' in self.magnet:
                self.magnet['mcvkey_shaft'] = self.magnet['mcvkey_mshaft']
            for mcv in ('mcvkey_yoke', 'mcvkey_shaft'):
                if mcv not in self.magnet:
                    self.magnet[mcv] = 'dummy'
        if 'coord_system' in parameters:
            self.move_action = 1
        else:
            self.coord_system = 0
            self.move_action = 0

        try:
            self.set_num_slots_gen()
        except (AttributeError):
            pass
        try:
            if 'wdgtype' not in self.windings:
                self.windings['wdgtype'] = 'SYM'
        except AttributeError:
            pass

        try:
            self.commutator = self.windings['wdgtype'] == 'CMM'
        except AttributeError:
            self.commutator = False
        try:
            self.windings['cufilfact'] = self.windings['fillfac']
        except (KeyError, AttributeError):
            pass

    def set_num_slots_gen(self):
        if 'num_slots_gen' not in self.stator:
            try:
                m = self.windings['num_phases']
            except KeyError:
                m = 1

            if hasattr(self, 'magnet'):
                slotgen = [m*self.poles]
            else:
                slotgen = [self.poles]
            try:
                slotgen.append(self.stator['num_slots'])
            except KeyError:
                pass

            try:
                slotgen.append(self.rotor['num_slots'])
            except (AttributeError, KeyError):
                pass

            g = np.gcd.reduce(slotgen)
            if hasattr(self, 'magnet') and g > 1:
                g /= m
            if 'num_slots' in self.stator:
                Q1 = self.stator['num_slots']
                self.stator['num_slots_gen'] = Q1 // g
            else:
                try:
                    Q2 = self.rotor['num_slots']
                    self.rotor['num_slots_gen'] = Q2 // g
                except (AttributeError, KeyError):
                    pass

    def set_mcvkey_magnet(self, mcvkey):
        self.mcvkey_magnet = mcvkey

    def get_mcvkey_magnet(self):
        try:
            return self.mcvkey_magnet
        except AttributeError:
            return ''

    def _set_magnet_material(self, comp, magcurves, magnetmat):
        names = []
        missing = []
        magnet = 0
        mcv = 0
        try:
            if magnetmat:
                magnet = magnetmat.find(comp['material'])
            if magnet and 'mcvkey' in magnet:
                if magcurves:
                    mcv = magcurves.find(magnet['mcvkey'])
                if mcv:
                    logger.debug('magnet mcv %s', mcv)
                    comp['mcvkey_magnet'] = mcv
                    names.append((mcv, 1.0))
                else:
                    missing.append(magnet['mcvkey'])
                    logger.error('magnet mcv %s not found',
                                 magnet['mcvkey'])
        except KeyError:
            pass
        except AttributeError:
            if 'material' in comp:
                missing.append(comp['material'])
                logger.error('magnet material %s not found',
                             comp['material'])
        return names, missing

    def set_magcurves(self, magcurves, magnetmat={}):
        """set and return real names of magnetizing curve material

        Args:
            magcurves: :class: 'MagnetizingCurve' including
                              magnetizing curve materials

        Return:
            set of magnetizing curve names with fillfac that have to be created

        """
        names = []
        missing = []

        mcv = 0
        if 'stator' in self.__dict__:
            for k in ('mcvkey_yoke', 'mcvkey_teeth'):
                mcvname = k+'_name'
                fillfac = self.stator.get('fillfac', 1.0)
                if mcvname not in self.stator:
                    try:
                        if self.stator[k] != 'dummy':
                            if magcurves:
                                mcv = magcurves.find(self.stator[k])
                            if mcv:
                                logger.debug('stator mcv %s', mcv)
                                self.stator[k] = magcurves.fix_name(
                                    mcv, fillfac)
                                names.append((mcv, fillfac))
                                self.stator[mcvname] = mcv
                            else:
                                missing.append(self.stator[k])
                                logger.error('stator mcv %s not found',
                                             self.stator[k])
                    except KeyError:
                        pass

                else:  # mcvname in self.stator:
                    names.append((self.stator[mcvname], fillfac))
            n, m = self._set_magnet_material(self.stator, magcurves, magnetmat)
            names += n
            missing += m

        if 'magnet' in self.__dict__:
            rotor = self.magnet
            subregion = 'magnet'
        elif 'rotor' in self.__dict__:
            rotor = self.rotor
            subregion = 'rotor'
        else:
            # no magnet no rotor
            return set(names)

        fillfac = rotor.get('fillfac', 1.0)
        try:
            if (rotor['mcvkey_yoke'] != 'dummy' and
                    'mcvkey_yoke_name' not in rotor):
                if magcurves:
                    mcv = magcurves.find(rotor['mcvkey_yoke'])
                if mcv:
                    logger.debug('%s mcv %s',
                                 subregion, mcv)
                    rotor['mcvkey_yoke'] = magcurves.fix_name(
                        mcv, fillfac)
                    rotor['mcvkey_yoke_name'] = mcv
                    names.append((mcv, fillfac))
                else:
                    missing.append(rotor['mcvkey_yoke'])
                    logger.error('%s mcv %s not found',
                                 subregion,
                                 rotor['mcvkey_yoke'])
            elif 'mcvkey_yoke_name' in rotor:
                names.append((rotor['mcvkey_yoke_name'], fillfac))

        except KeyError:
            pass

        try:
            if (rotor['mcvkey_shaft'] != 'dummy' and
                    'mcvkey_shaft_name' not in rotor):
                if magcurves:
                    mcv = magcurves.find(rotor['mcvkey_shaft'])
                if mcv:
                    logger.debug('shaft mcv %s', mcv)
                    rotor['mcvkey_shaft'] = magcurves.fix_name(mcv)
                    self.stator['mcvkey_shaft_name'] = mcv
                    names.append((mcv, 1.0))
                else:
                    missing.append(rotor['mcvkey_shaft'])
                    logger.error(' shaft %s not found',
                                 rotor['mcvkey_shaft'])
            elif 'mcvkey_shaft_name' in rotor:
                names.append((self.stator['mcvkey_shaft_name'], 1.0))

        except KeyError:
            pass

        if 'magnet' in self.__dict__:
            n, m = self._set_magnet_material(self.magnet, magcurves, magnetmat)
            missing += m
            names += n
        if missing:
            raise MCerror("MC pars missing: {}".format(
                ', '.join(set(missing))))
        return set(names)

    def statortype(self):
        """return type of stator slot"""
        for k in self.stator:
            if isinstance(self.stator[k], dict):
                return k
        raise AttributeError("Missing stator slot model in {}".format(
            self.stator))

    def rotortype(self):
        """return type of rotor slot"""
        for k in self.rotor:
            if isinstance(self.rotor[k], dict):
                return k
        raise AttributeError("Missing rotor model in {}".format(self.magnet))

    def magnettype(self):
        """return type of magnet slot"""
        for k in self.magnet:
            if k != 'material' and isinstance(self.magnet[k], dict):
                return k
        raise AttributeError("Missing magnet model in {}".format(self.magnet))

    def is_complete(self):
        """check completeness of models"""
        if self.is_dxffile():
            return True
        try:
            self.statortype()
            if hasattr(self, 'rotor'):
                self.rotortype()
            else:
                self.magnettype()
            return True
        except AttributeError:
            return False

    def is_dxffile(self):
        if 'dxffile' in self.__dict__:
            if isinstance(self.dxffile, dict):
                return True
        return False

    def has_magnet(self):  # either rotor or magnet
        return hasattr(self, 'magnet')


class FeaModel(Model):
    def __init__(self, parameters):
        self.cufilfact = 0.45
        self.culength = 1.4
        self.wind_temp = 20
        self.slot_indul = 0.0
        self.skew_angle = 0.0
        self.num_skew_steps = 0
        self.num_par_wdgs = 1
        self.eval_force = 0
        self.optim_i_up = 0
        self.plots = []
        self.airgap_induc = []
        super(self.__class__, self).__init__(parameters)



"""
  GT-ISE Femag API

  induction machine (electrical circuit model)

  Copyright 2021: Semafor Informatik & Energie AG, Switzerland
"""
import numpy as np
import scipy.optimize as so
import logging

EPS = 1e-13
KTH = 0.003921  # temperature coefficient of resistance
TREF = 20.0  # reference temperature of resistance

eecdefaults = dict(
    zeta1=0.2,
    zeta2=1.6,
    gam=0.7,
    kh=2,
    tcu1=20,
    rotor_mass=0,
    kfric_b=1,
    pl2v=0.5,
    tcu2=20)

logger = logging.getLogger('im')


def xiskin(w, temp, zeta):
    return zeta*np.sqrt(abs(w)/(2*np.pi)/(50*(1+KTH*(temp-TREF))))


def kskinl(xi, nl):
    if abs(xi) < EPS:
        return 1.0
    xi2 = 2*xi
    nl2 = nl*nl
    return 3 / (nl2*xi2)*(np.sinh(xi2) - np.sin(xi2)) / \
        (np.cosh(xi2)-np.cos(xi2)) + \
        ((nl2-1)/(nl2*xi)*(np.sinh(xi)+np.sin(xi)) /
            (np.cosh(xi)+np.cos(xi)))


def kskinr(xi, nl):
    if abs(xi) < EPS:
        return 1.0
    xi2 = 2*xi
    nl2 = nl*nl
    return xi*((np.sinh(xi2)+np.sin(xi2))/(np.cosh(xi2)-np.cos(xi2))) + \
        ((nl2-1) / 3 * xi2*((np.sinh(xi)-np.sin(xi)) /
                            (np.cosh(xi)+np.cos(xi))))


def resistance(r0, w, temp, zeta, gam, nh):
    xi = xiskin(w, temp, zeta)
    return r0*(1.+KTH*(temp - TREF))*(gam + kskinr(xi, nh)) / (1. + gam)


class Component:
    def __init__(self, arg):
        """initialize this object either from a json file
        or from a set of parameters"""
        parameters = arg
        if type(arg) == str:
            file = open(arg)
            parameters = json.load(file)
            file.close()
        for k in parameters.keys():
            setattr(self, k, parameters[k])

    def to_json(self):
        """return json format of this object as string"""
        return json.dumps(self.__dict__)

    def __str__(self):
        "return string format of this object"
        return repr(self.__dict__)

    def __repr__(self):
        "representation of this object"
        return self.__str__()


class InductionMachine(Component):
    """basic induction machine model with T equivalent circuit"""

    def __init__(self, parameters):
        Component.__init__(self, parameters)
        if 'f1ref' in parameters:
            self.wref = 2*np.pi*self.f1ref
        if 'u1ref' in parameters and 'f1ref' in parameters:
            self.psiref = self.u1ref/self.wref
        if 'lh' in parameters:
            self.ims = 0
            self.psiref = 1
            self.iml = 1/self.lh
            self.mexp = 0
        for k in eecdefaults.keys():
            if not hasattr(self, k):
                setattr(self, k, eecdefaults[k])

    def rfe(self, w, psi):
        """equivalent resistance for iron losses"""
        try:
            if w == 0:
                return 0
            else:
                return self.m*((w*w)*(psi * psi)) / (
                    (self.fec + self.fee *
                     np.power(np.abs(psi)/self.psiref, self.fexp)) *
                    (0.75 + 0.25*(np.abs(w)/self.wref)) *
                    (np.abs(w)/self.wref) *
                    np.power((np.abs(psi)/self.psiref), 2.0))
        except AttributeError:
            pass
        return self.rh

    def imag(self, psi):
        """magnetizing current"""
        return (self.iml * np.abs(psi)/self.psiref +
                self.ims*np.power(np.abs(psi)/self.psiref, self.mexp))
        # return psi/self.lh

    def lrot(self, w):
        """rotor leakage inductivity"""
        kl2 = 1.+self.pl2v*(kskinl(
            xiskin(w, self.tcu2, self.zeta2), 1)-1)
        return self.lsigma2*kl2

    def lstat(self, w):
        """stator leakage inductivity"""
        return self.lsigma1

    def rrot(self, w):
        """rotor resistance"""
        return resistance(self.r2, w, self.tcu2, self.zeta2,
                          0.0, 1)

    def rstat(self, w):
        """stator resistance"""
        return resistance(self.r1, w, self.tcu1, self.zeta1,
                          self.gam, self.kh)
        # return self.r1

    def sigma(self, w, psi):
        """leakage coefficient"""
        lh = psi / self.imag(psi)
        return (1 - (np.power(np.abs(lh), 2.0)) /
                ((lh+self.lstat(w))*(lh+self.lrot(w))))

    def u1(self, w1, psi, wm):
        """stator voltage"""
        istat = self.i1(w1, psi, wm)
        z1 = (self.rstat(w1) + w1*self.lstat(w1)*1j)
        return w1*psi*1j + istat*z1

    def i1(self, w1, psi, wm):
        """stator current"""
        imag = self.imag(psi)
        if abs(w1) > 0:
            imag += w1*psi/self.rfe(w1, psi)
        return self.i2(w1, psi, wm) + imag

    def i2(self, w1, psi, wm):
        """rotor current"""
        w2 = w1 - self.p * wm
        if abs(w2) > EPS:
            z2 = (self.rrot(w2) + w2*self.lrot(w2)*1j)
            return w2*psi*1j/z2
        return 0

    def w1torque(self, w1, u1max, psi, wm, tload):
        """calculate difference between load and motor torque"""
        # check stator voltage
        u1 = self.u1(w1, psi, wm)
        self.psi = psi
        if abs(u1) > u1max:  # must adjust flux
            self.psi = so.bisect(
                lambda psi: u1max - abs(self.u1(w1, psi, wm)),
                psi, 0.1*psi)
        return self.torque(w1, self.psi, wm)-tload

    def pullouttorque(self, w1, u1):
        """pull out torque"""
        sk = self.sk(w1, u1/w1)
        w2 = sk*w1
        r2 = self.rrot(w2)
        psi = u1/w1
        lh = psi/self.imag(psi)
        x1 = w1*(lh+self.lstat(w1))
        x2 = w1*(lh+self.lrot(w2))
        r1 = self.rstat(w1)
        sigma = self.sigma(w1, u1/w1)
        return self.m*self.p * u1**2/w1 * (1 - sigma) / \
            ((r1**2 + x1**2)*r2/(sk * x1 * x2) +
             sk * x2*(r2**2 + sigma**2 * x1**2)/(r2 * x1) + 2*r1*(1-sigma))

    def sk(self, w1, psi):
        """pullout slip"""
        r2 = self.rrot(0.)
        lh = psi/self.imag(psi)
        x1 = w1*(lh+self.lstat(w1))
        x2 = w1*(lh+self.lrot(0.))
        r1 = self.rstat(w1)
        return r2/x2*np.sqrt((r1**2+x1**2)/(
            self.sigma(w1, psi)**2*x1**2 + r1**2))

    def torque(self, w1, psi, wm):
        """electric torque (in airgap)"""
        w2 = w1-self.p*wm
        if w2 == 0:
            return 0.
        r2 = self.rrot(w2)
        i2 = self.i2(w1, psi, wm)
        return self.m*self.p/w2*r2*(i2*i2.conjugate()).real

    def torqueu(self, w1, u1max, wm):
        """electric torque (in airgap)"""
        w2 = w1-self.p*wm
        if w2 == 0:
            return 0.
        # find psi
        psi = u1max/w1
        self.psi = so.fsolve(
            lambda psi: u1max - abs(self.u1(w1, psi, wm)),
            psi)[0]
        return self.torque(w1, self.psi, wm)

    def w1(self, u1max, psi, tload, wm):
        """calculate stator frequency with given torque and speed"""
        wsync = max(wm*self.p, 0.001)
        sk = self.rrot(0.)/(wsync*(self.lstat(wsync) + self.lrot(0.0)))
        if tload < 0:
            a, b = wsync, wsync*(1-sk)
        else:
            a, b = wsync, wsync*(1+sk)
        try:
            return so.bisect(self.w1torque, a, b,
                             args=(u1max, psi, wm, tload))
        except ValueError as ex:
            logger.debug("a, b: %s",
                         (a, b, tload,
                          self.torque(a, self.psi, wm),
                          self.torque(b, self.psi, wm)))
            raise ex

    def wmfweak(self, u1max, psi, torque):
        """return lower speed limit of field weakening range"""
        wm0 = u1max/psi/self.p
        try:
            return so.fsolve(
                lambda wm: u1max - np.abs(self.u1(
                    self.w1(u1max, psi, torque, wm), psi, wm)),
                wm0)[0]
        except ValueError as ex:
            logger.error("wm0 %s, umax: %s", (wm0, 0.9*wm0),
                         (u1max, np.abs(self.u1(
                             self.w1(u1max, psi, torque, wm0), psi, wm0)),
                          np.abs(self.u1(
                              self.w1(u1max, psi, torque, wm0), psi, wm0))))
            raise ex

    def characteristics(self, T, n, u1max, nsamples=50):
        """calculate torque speed characteristics.
        return dict with list values of
        id, iq, n, T, ud, uq, u1, i1,
        beta, gamma, phi, cosphi, pmech, n_type

        Keyword arguments:
        T -- (float) the maximum torque in Nm
        n -- (float) the maximum speed in 1/s
        u1max -- (float) the maximum voltage in V rms
        nsamples -- (optional) number of speed samples
        """

        wmType = self.wmfweak(u1max, self.psiref, T)
        pmmax = wmType*T
        KPO = 0.9
        wmPullout = so.fsolve(
            lambda wx: (self.pullouttorque(self.p*wx, u1max) - abs(pmmax/wx)),
            wmType)
        wmtab0 = np.linspace(wmType, 2*wmPullout[0])
        for wm, tq in zip(wmtab0, [pmmax/wx for wx in wmtab0]):
            logger.debug("u1 %g psi %g tq %g wm %g",
                         u1max, self.psiref, tq, wm)
            try:
                w1 = self.w1(u1max, self.psiref, tq, wm)
            except ValueError:
                wpo = KPO * pmmax/tq
                break

        wmMax = 1.5*wpo
        if n:
            wmMax = 2*np.pi*n
        wmtab0 = np.linspace(wpo, wmMax)
        for wm, tq in zip(wmtab0, [pmmax/wx**2 for wx in wmtab0]):
            logger.debug("u1 %g psi %g tq %g wm %g",
                         u1max, self.psiref, tq, wm)
            try:
                w1 = self.w1(u1max, self.psiref, tq, wm)
            except ValueError:
                wmMax = wm
                break

        logger.info("wmtype %f wpo %f wmmax %f", wmType, wpo, wmMax)

        if wmType < wpo < wmMax:
            wmtab = []
            dw = 0
            wmrange = sorted([0, wmType, wpo, wmMax])
            for a, b in zip(wmrange, wmrange[1:]):
                nx = int(round(nsamples*(b-a)/wmrange[-1]))
                dw = (b-a)/nx
                wmtab += np.linspace(a+dw, b, nx).tolist()
            wmtab[0] = 0
        elif wmMax < wmType:
            wmrange = sorted([0, wmMax])
            wmtab = np.linspace(0, wmMax, nsamples).tolist()
        else:
            wmrange = sorted([0, wmType, wmMax])
            nx = int(round(nsamples*wmType/wmMax))
            dw = (wmMax-wmType)/(nsamples-nx)
            wmtab = (np.linspace(0, wmType, nx).tolist() +
                     np.linspace(wmType+dw, wmMax, nsamples-nx).tolist())

        logger.info("Speed range %s", wmrange)

        def tload2(wm):
            if wm < wmType and wm < wpo:
                return T
            if wm < wpo:
                return pmmax/wm
            return wpo*pmmax/wm**2

        r = dict(u1=[], i1=[], T=[], cosphi=[], n=[],
                 plfe1=[], plcu1=[], plcu2=[])
        T = [tload2(wx) for wx in wmtab]
        tfric = self.kfric_b*self.rotor_mass*30e-3/np.pi
        w1tab = []
        for wm, tq in zip(wmtab, T):
            logger.debug("u1 %g psi %g tq %g wm %g",
                         u1max, self.psiref, tq, wm)
            try:
                w1 = self.w1(u1max, self.psiref, tq, wm)
                w1tab.append(w1)
                u1 = self.u1(w1, self.psi, wm)
                r['u1'].append(np.abs(u1))
                i1 = self.i1(w1, self.psi, wm)
                r['i1'].append(np.abs(i1))
                r['cosphi'].append(np.cos(np.angle(u1) - np.angle(i1)))
                r['plfe1'].append(self.m*np.abs(u1)**2/self.rfe(w1, self.psi))
                i2 = self.i2(w1, self.psi, wm)
                r['plcu1'].append(self.m*np.abs(i1)**2*self.rstat(w1))
                r['plcu2'].append(self.m*np.abs(i2)**2*self.rrot(w1-self.p*wm))
                r['T'].append(tq - tfric)
                r['n'].append(wm/2/np.pi)
            except ValueError:
                break
        r['plfric'] = [2*np.pi*n*tfric for n in r['n']]
        r['pmech'] = [2*np.pi*n*tq for n, tq in zip(r['n'], r['T'])]
        pmech = np.array(r['pmech'])
        pltotal = (np.array(r['plfe1']) + np.array(r['plfric']) +
                   np.array(r['plcu1']) + np.array(r['plcu2']))
        r['losses'] = pltotal.tolist()

        if pmech.any():
            p1 = pmech + pltotal
            if np.abs(pmech[0]) < 1e-12:
                r['eta'] = [0]
                i = 1
            else:
                r['eta'] = []
                i = 0
            if np.all(abs(p1[i:]) > abs(pmech[i:])):
                r['eta'] += (pmech[i:]/(p1[i:])).tolist()
            else:
                r['eta'] += (p1[i:]/pmech[i:]).tolist()

        return r


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import femagtools.plot

    mpars = dict(
        u1type=2184/np.sqrt(3),
        f1type=53,
        m=3,
        p=2,
        rh=524.2776,
        # lh=35.88343e-3,
        r1=0.043,
        r2=0.0346,
        lsigma1=1.15e-3,
        lsigma2=0.73361e-3,
        fec=9100,
        fee=0,
        fexp=7.0,
        pfric=2544.6,
        rexp=2.4,
        iml=79.8286,
        ims=27.5346,
        mexp=6.5)

    im = InductionMachine(mpars)
    torque = 6543
    u1max = 2184/np.sqrt(3)
    r = im.characteristics(torque, 0, u1max)
    # plt.plot(np.array(r['n'])*60, r['cosphi'])
    #    plt.plot(np.array(r['n'])*60, r['plcu'])
    #    plt.plot(np.array(r['n'])*60, r['losses'])
    femagtools.plot.characteristics(r)
    plt.show()



# -*- coding: utf-8 -*-
"""
    femagtools.job
    ~~~~~~~~~~~~~~

    Creating and managing calculation jobs.



"""
import os
import platform
import shutil
import glob
import femagtools.bch
import femagtools.config as cfg
import logging
import uuid
import importlib

logger = logging.getLogger(__name__)

# lines that contains this words are detected as error messages
protfile_error = ['Model has no field sources']
femagfile_error = ['ms_error',
                   'unexpected terminal input in quiet mode',
                   'Fortran runtime error',
                   'Exception system errors']

# https://python-3-patterns-idioms-test.readthedocs.io/en/latest/Factory.html


class TaskFactory:
    factories = {}

    def addFactory(typename, taskFactory):
        TaskFactory.factories.put[typename] = taskFactory
    addFactory = staticmethod(addFactory)

    def createTask(typename, taskid, dir, result_func):
        if typename not in TaskFactory.factories:
            try:
                TaskFactory.factories[typename] = \
                    eval(typename + '.Factory()')
            except NameError:
                modname, classname = typename.split('.')
                mod = importlib.import_module(modname)
                TaskFactory.factories[typename] = getattr(mod,
                                                          classname).Factory()
        return TaskFactory.factories[typename].create(taskid, dir, result_func)
    createTask = staticmethod(createTask)


class Task(object):
    """represents a single execution unit that may include data files"""

    def __init__(self, id, directory, result_func=None):
        self.result_func = result_func
        self.directory = directory
        try:
            os.makedirs(self.directory)
        except OSError as e:
            if e.errno != 17:
                e.args += self.directory
                raise
            pass
        self.transfer_files = []
        self.status = None
        self.fsl_file = None
        self.id = id

    def add_file(self, fname, content=None):
        """adds a file required by this task

        Args:
            fname: file name
            content: list of str written to file if not None"""
        base = os.path.basename(fname)
        self.transfer_files.append(fname)
        if os.path.splitext(base)[-1] == '.fsl':
            self.fsl_file = base

        if content is None:
            dest = os.path.join(self.directory, base)
            if not os.access(dest, os.R_OK):
                shutil.copy(fname, dest)
            return

        # this file has to be created
        with open(os.path.join(self.directory, base), 'w') as f:
            f.writelines('\n'.join(content))

    def get_results(self):
        """returns result of most recent BCH file (or project specific results if result_func is set)"""
        if self.result_func:
            return self.result_func(self)

        # read latest bch file if any
        bchfile_list = sorted(glob.glob(os.path.join(
            self.directory, '*_[0-9][0-9][0-9].B*CH')))
        if bchfile_list:
            result = femagtools.bch.Reader()
            with open(bchfile_list[-1]) as f:
                logger.info("Reading %s",
                            bchfile_list[-1])
                result.read(f)
            return result
            # logger.info("%s %s", result.version, result.type),
        else:
            asm_list = sorted(glob.glob(os.path.join(
                self.directory, '*_[0-9][0-9][0-9].ASM')))
            if asm_list:
                logger.info("Reading %s",
                            asm_list[-1])
                return femagtools.asm.read(asm_list[-1])
            msg = 'no BCH (or ASM) files in {}'.format(self.directory)
            logger.error(msg)
        result = dict(error=msg)
        return result

    def readErrorMessage(self, html=True):
        errstr = ""
        if html:
            sepIn = '<p><b>{}</b><br/>'
            sepOut = '</p>'
            newLine = '<br/>'
        else:
            sepIn = '=== {} ===\n'
            sepOut = '\n\n'
            newLine = '\n'

        errstr += sepIn.format(self.directory)

        # read latest prot file
        protfile_list = sorted(
            glob.glob(os.path.join(self.directory, '*.PROT')))
        if protfile_list:
            with open(protfile_list[-1], encoding='latin1') as f:
                e = [em in l for em in protfile_error for l in f.readlines()]
                if sum(e):
                    logger.info("Prot Files Message.\n '%s'\n\nDirectory '%s'.",
                                e, self.directory)
                    errstr += "ProtFile: {}{}".format(e, newLine)

        # read femag.err file
        try:
            with open(os.path.join(self.directory, 'femag.err'), 'r') as file:
                e = [em in l for em in femagfile_error for l in file.readlines()]
                if sum(e):
                    logger.error("Femag Error detected\n '%s'\n\nDirectory '%s'.",
                                 e, self.directory)
                    errstr += "{}{}".format(e, newLine)
        except Exception as e:
            logger.debug("Exception: %s", e)
            pass
        errstr += sepOut
        return errstr

    def __eq__(self, other):
        return isinstance(other, type(self)) and \
            self.directory == other.directory

    class Factory:
        def create(self, id, dir, result_func): return Task(
            id, dir, result_func)


class CloudTask(Task):
    def __init__(self, id, directory, result_func=None):
        super(self.__class__, self).__init__(id, directory, result_func)
        import tarfile
        self.file = "{}.tar.gz".format(self.directory)
        self.tar_file = tarfile.open(self.file, "w:gz")
        # Used for amazon
        self.ec2_instance = None

    def add_file(self, fname, content=None):
        base = os.path.basename(fname)
        self.transfer_files.append(fname)
        if os.path.splitext(base)[-1] == '.fsl':
            self.fsl_file = base

        info = self.tar_file.tarinfo()
        info.name = base
        if content is None:
            info.size = os.path.getsize(fname)
            self.tar_file.addfile(info, open(fname, 'rb'))
            return

        import io
        if type(content) is list:
            content = '\n'.join(content)
        info.size = len(content)
        data = io.BytesIO(str.encode(content))
        self.tar_file.addfile(info, data)

    class Factory:
        def create(self, id, dir, result_func): return CloudTask(
            id, dir, result_func)


class Job(object):
    """represents a FEMAG job consisting of one or more tasks
    each to be executed by a dedicated process
    """

    def __init__(self, basedir):
        self.runDirPrefix = ''
        self.basedir = basedir
        self.tasks = []

    def cleanup(self):
        """removes all task directories of previous run"""
        for task in self.tasks:
            logger.debug("rm %s", task.directory)
            shutil.rmtree(task.directory, ignore_errors=True)
        self.tasks = []

    def add_task(self, result_func=None):
        "adds a new task to this job"
        taskid = "{}-{}".format(
            str(uuid.uuid4()).split('-')[0], len(self.tasks))
        dir = os.path.join(self.basedir,
                           '{}{:d}'.format(self.runDirPrefix,
                                           len(self.tasks)))
        t = TaskFactory.createTask('Task', taskid, dir, result_func)
        self.tasks.append(t)
        return t

    def setExitStatus(self, taskid, status):
        "set exit status of task"
        self.tasks[taskid].status = status

    def get_results(self):
        for t in self.tasks:
            yield t.get_results()


class CondorJob(Job):
    """represents a femag job that is to be run in HT Condor"""

    def __init__(self, basedir):
        super(self.__class__, self).__init__(basedir)

    def prepareDescription(self):
        # create a flatten list of all files to be transferred
        transfer_files = [item for sublist in
                          [t.transfer_files for t in self.tasks]
                          for item in sublist]

        OpSys = "WINDOWS" if platform.system() == "Windows" else "LINUX"

        fslFilename = ''
        for f in transfer_files:
            b, ext = os.path.splitext(f)
            if ext == '.fsl':
                fslFilename = f
                break
        numRuns = len(self.tasks)
# $(Process)                os.path.join(self.basedir, self.runDirPrefix)),
        submit = [
            'InitialDir   = {}/{}$(Process)'.format(
                self.basedir, self.runDirPrefix),
            'Universe     = vanilla',
            'Executable   = {}'.format(cfg.get_executable()),
            'Arguments    = -b {}'.format(fslFilename),
            'Output       = femag.out',
            'Log          = femag.log',
            'Error        = femag.err',
            'Notification = never',
            'input        = /dev/null',
            'transfer_input_files = {}'.format(','.join(set(transfer_files))),
            'requirements = (OpSys=="{}") && Arch=="x86_64"'.format(OpSys),
            'Queue {}'.format(numRuns), '']
        filename = os.path.join(self.basedir, "femag.submit")
        with open(os.path.join(self.basedir,
                               "femag.submit"), 'w') as submitFile:
            submitFile.writelines('\n'.join(submit))
        return filename


class CloudJob(Job):
    """Inheritance of :py:class:`Job`

    Represents a femag amazon job"""

    def __init__(self, basedir):
        super(self.__class__, self).__init__(basedir)

    def add_task(self, result_func=None):
        "adds a new :py:class:`CloudTask` to this job"
        taskid = "{}-{}".format(str(uuid.uuid4()), len(self.tasks))
        dir = os.path.join(self.basedir,
                           '{}{:d}'.format(self.runDirPrefix,
                                           len(self.tasks)))
        t = TaskFactory.createTask('CloudTask', taskid, dir, result_func)
        self.tasks.append(t)
        return t



#
# Upload to pypi:
#   python setup.py sdist bdist_wheel
#   twine upload dist/*
#
#   To use these commands install wheel and twine
#   pip install wheel twine
#
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
import os
import io
import itertools
import re
import ast

_version_re = re.compile(r'__version__\s+=\s+(.*)')
_license_re = re.compile(r'__license__\s+=\s+(.*)')
_author_re = re.compile(r'__author__\s+=\s+(.*)')
line_numbers = 20

with io.open('femagtools/__init__.py', 'r') as f:
    meta_data = list(itertools.islice(f, line_numbers))
meta_data = ''.join(meta_data)

version = str(ast.literal_eval(_version_re.search(meta_data).group(1)))
license = str(ast.literal_eval(_license_re.search(meta_data).group(1)))
author = str(ast.literal_eval(_author_re.search(meta_data).group(1)))

here = os.path.abspath(os.path.dirname(__file__))
# Get the long description from the README file
with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    description='Python API for FEMAG',
    long_description=long_description,
    author=author,
    url='https://github.com/SEMAFORInformatik/femagtools',
    author_email='tar@semafor.ch',
    version=version,
    platforms="any",
    install_requires=['numpy<=1.20.2', 'scipy>=1.7', 'mako', 'six', 'lmfit',
                      'dxfgrabber', 'networkx', 'netCDF4'],
    extras_require={"meshio": ["meshio<=4.4.6", "lxml"]},
    packages=['femagtools', 'femagtools.moo', 'femagtools.dxfsl'],
    package_data={'femagtools': ['templates/*.mako']},
    license=license,
    name='femagtools',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering'],
    entry_points={"console_scripts": [
        "femagtools-plot = femagtools.plot:main",
        "femagtools-convert = femagtools.convert:main",
        "femagtools-bchxml = femagtools.bchxml:main",
        "femagtools-dxfsl = femagtools.dxfsl.conv:main"]}
)



"""
  femagtools.dxfsl.converter
  ~~~~~~~~~~~~~~~~~~~~~~~~~~

  read a dxf file and create a plot or fsl file

 Authors: Ronald Tanner, Beat Holm
"""
import os
from femagtools.dxfsl.geom import Geometry, dxfshapes, femshapes
from femagtools.dxfsl.shape import Shape
from femagtools.dxfsl.fslrenderer import FslRenderer, agndst
from femagtools.dxfsl.plotrenderer import PlotRenderer
import logging
import logging.config
import numpy as np
import sys

logger = logging.getLogger(__name__)


def symmetry_search(machine,
                    plt,  # plotter
                    kind,
                    symtol=0.0,
                    sympart=0,
                    is_inner=False,
                    is_outer=False,
                    show_plots=True,
                    debug_mode=False,
                    rows=1,
                    cols=1,
                    num=1):
    logger.info("symmetry search for %s", kind)

    machine.clear_cut_lines()
    if show_plots and debug_mode:
        plt.render_elements(machine.geom, Shape,
                            neighbors=True, title=kind)

    if sympart > 0:
        if not machine.is_full():
            logger.error("force symmetry failed")
            sys.exit(1)
        machine_ok = machine.get_forced_symmetry(sympart)
        machine_ok.set_minmax_radius()
        machine_ok.create_auxiliary_lines()
        machine_ok.set_kind(kind)
        logger.info("*** End of symmetry search for %s ***", kind)
        return machine_ok

    if not machine.find_symmetry(symtol):
        logger.info(" - {}: no symmetry axis found".format(kind))
        if show_plots:
            plt.add_emptyplot(rows, cols, num, 'no symmetry axis')

        machine_mirror = machine.get_symmetry_mirror()
        machine_slice = machine
    else:
        logger.info(" - {}: symmetry axis found !!".format(kind))
        if show_plots:
            plt.render_elements(machine.geom, Shape,
                                title=kind+' (symmetrylines)',
                                draw_inside=True,
                                rows=rows, cols=cols, num=num, show=False)
        machine_slice = machine.get_symmetry_slice()
        if machine_slice is None:
            machine.kind = kind
            logger.info(" - no slice extracted ?!?")
            return machine

        machine_mirror = machine_slice.get_symmetry_mirror()

    if machine_mirror is None:
        logger.info(" - no mirror found")
        if not machine_slice.is_startangle_zero():
            machine_slice.rotate_to(0.0)
            machine_slice.set_alfa_and_corners()

        machine_ok = machine_slice
    else:
        if show_plots and debug_mode:
            plt.render_elements(machine_mirror.mirror_geom, Shape,
                                title='Mirror of '+kind,
                                rows=rows, cols=cols, num=num, show=True)

        logger.info(" - mirror found")
        machine_next_mirror = machine_mirror.get_symmetry_mirror()
        while machine_next_mirror is not None:
            logger.info(" - another mirror found")
            machine_mirror = machine_next_mirror
            machine_next_mirror = machine_mirror.get_symmetry_mirror()

        machine_ok = machine_mirror

    machine_ok.set_minmax_radius()
    # machine_ok.complete_hull(is_inner, is_outer)
    machine_ok.create_auxiliary_lines()
    machine_ok.set_kind(kind)

    logger.info("*** End of symmetry search for %s ***", kind)
    return machine_ok


def convert(dxfile,
            rtol=1e-03,
            atol=0.005,
            mindist=0.0,
            symtol=0.001,
            sympart=0,
            split=False,
            inner_name='inner',
            outer_name='outer',
            part=(),
            airgap=0.0,
            airgap2=0.0,
            da=0.0,
            dy=0.0,
            nodedist=1,
            view_only=False,
            view_korr=False,
            show_plots=False,
            show_areas=False,
            write_fsl=True,
            write_png=False,
            write_id=False,
            debug_mode=False):
    layers = ()
    conv = {}

    basename = os.path.basename(dxfile).split('.')[0]
    logger.info("start processing %s", basename)

    if part:
        if part[0] not in ('rotor', 'stator'):
            logger.error('FATAL: Parameter rotor or stator expected')
            return dict(error='unknown part {}'.format(part))
        if part[1] not in ('in', 'out'):
            logger.error('"{}" has to be defined in/out'.format(part[0]))
            return dict(error='unknown location {}'.format(part[1]))
    else:
        if da:
            logger.warn("distance airgap (da) ignored")
            da = 0.0
        if dy:
            logger.warn("distance yoke (dy) ignored")
            dy = 0.0

    try:
        if dxfile.split('.')[-1] == 'fem':
            basegeom = Geometry(femshapes(dxfile),
                                rtol=rtol,
                                atol=atol,
                                split=split)
        else:
            basegeom = Geometry(dxfshapes(dxfile,
                                          mindist=mindist,
                                          layers=layers),
                                rtol=rtol,
                                atol=atol,
                                split=split)
    except FileNotFoundError as ex:
        logger.error(ex)
        return dict()

    logger.info("total elements %s", len(basegeom.g.edges()))

    p = PlotRenderer()

    if view_only:
        logger.info("View only")
        if view_korr:
            logger.info("With Corrections")
            basegeom.search_all_overlapping_elements()
            basegeom.search_all_appendices()

        p.render_elements(basegeom, Shape,
                          neighbors=True,
                          png=write_png,
                          show=True)
        return dict()

    basegeom.search_all_overlapping_elements()

    machine_base = basegeom.get_machine()
    if show_plots:
        p.render_elements(basegeom, Shape,
                          title=os.path.basename(dxfile),
                          with_hull=False,
                          rows=3, cols=2, num=1, show=debug_mode)

    basegeom.search_all_appendices()

    if not machine_base.is_a_machine():
        logger.warn("it's Not a Machine!!")
        return dict(error='machine not detected')

    if not (machine_base.part > 0):
        # machine shape is unclear
        machine_base.set_center(0.0, 0.0)
        machine_base.set_radius(9999999)

    machine = machine_base

    if machine.part_of_circle() == 0:
        logger.warn("No arc segment found")
        return dict(error='no arc segment found')

    machine.clear_cut_lines()
    machine.move_to_middle()
    if show_plots and debug_mode:
        p.render_elements(machine.geom, Shape,
                          title='Areas',
                          neighbors=True,
                          with_corners=False, show=True)

    if machine.airgap(airgap, airgap2, symtol):
        p.render_elements(machine.geom, Shape,
                          title='Search for airgap failed',
                          with_corners=False, show=True)
        return dict(error='no airgap found')

    if show_plots:
        p.render_elements(basegeom, Shape, neighbors=True,
                          title='Original with nodes',
                          rows=3, cols=2, num=2, show=False)

    machine.repair_hull()
    machine.geom.delete_all_appendices()

    if machine.has_airgap():
        machine_inner = machine.copy(0.0, 2*np.pi, True, True)
        machine_inner = symmetry_search(machine_inner,
                                        p,  # plot
                                        inner_name,
                                        is_inner=True,
                                        symtol=symtol,
                                        show_plots=show_plots,
                                        rows=3,  # rows
                                        cols=2,  # columns
                                        num=3)   # start num
        machine_inner.set_inner()

        machine_outer = machine.copy(0.0, 2*np.pi, True, False)
        machine_outer = symmetry_search(machine_outer,
                                        p,  # plot
                                        outer_name,
                                        is_outer=True,
                                        symtol=symtol,
                                        show_plots=show_plots,
                                        rows=3,  # rows
                                        cols=2,  # columns
                                        num=4)   # start num

        machine_inner.sync_with_counterpart(machine_outer)

        machine_inner.search_subregions()
        machine_outer.search_subregions()

        if machine_inner.has_mirrored_windings():
            logger.info("undo mirrored windings of %s", inner_name)
            machine_inner = machine_inner.undo_mirror()
            machine_inner.sync_with_counterpart(machine_outer)
            machine_inner.search_subregions()
            machine_inner.create_mirror_lines_outside_windings()

        elif machine_outer.has_mirrored_windings():
            logger.info("undo mirrored windings of %s", outer_name)
            machine_outer = machine_outer.undo_mirror()
            machine_inner.sync_with_counterpart(machine_outer)
            machine_outer.search_subregions()
            machine_outer.create_mirror_lines_outside_windings()

        machine_inner.delete_tiny_elements(mindist)
        machine_outer.delete_tiny_elements(mindist)
        logger.info("END of work: %s", basename)

        if show_plots:
            p.render_elements(machine_inner.geom, Shape,
                              draw_inside=True, title=inner_name,
                              rows=3, cols=2, num=5, show=False,
                              # with_nodes=True,
                              # neighbors=True,
                              write_id=write_id,
                              fill_areas=True)

            p.render_elements(machine_outer.geom, Shape,
                              draw_inside=True, title=outer_name,
                              rows=3, cols=2, num=6, show=False,
                              # with_nodes=True,
                              # neighbors=True,
                              write_id=write_id,
                              fill_areas=True)
            if write_png:
                p.write_plot(basename)
            else:
                p.show_plot()

        if show_areas:
            p.render_elements(machine_inner.geom, Shape,
                              title=inner_name,
                              show=True,
                              draw_inside=True,
                              neighbors=True,
                              fill_areas=True)
            p.render_areas(machine_inner.geom,
                           title=inner_name,
                           with_nodes=True,
                           single_view=True)

            p.render_elements(machine_outer.geom, Shape,
                              title=outer_name,
                              show=True,
                              draw_inside=True,
                              neighbors=True,
                              fill_areas=True)
            p.render_areas(machine_outer.geom,
                           title=outer_name,
                           with_nodes=True,
                           single_view=True)

        if write_fsl:
            if machine_inner.is_full() or machine_outer.is_full():
                logger.warning("it's not possible to create fsl-file")
                return None

            fslrenderer = FslRenderer(basename)
            inner = fslrenderer.render(machine_inner, inner=True)
            outer = fslrenderer.render(machine_outer, outer=True)

            if machine_inner.geom.is_rotor():
                conv['fsl_magnet'] = inner
                conv['fsl_stator'] = outer
            else:
                conv['fsl_magnet'] = inner
                conv['fsl_rotor'] = outer

            params = create_femag_parameters(machine_inner,
                                             machine_outer,
                                             nodedist)

            conv.update(params)
            conv['fsl'] = fslrenderer.render_main(
                machine,
                machine_inner, machine_outer,
                inner, outer,
                params)
    else:
        # No airgap found. This must be an inner or outer part
        name = "No_Airgap"
        inner = False
        outer = False
        params = None

        if part:
            if part[1] == 'in':
                name = inner_name
                inner = True
            else:
                name = outer_name
                outer = True

        machine = symmetry_search(machine,
                                  p,  # plot
                                  name,
                                  symtol=symtol,
                                  sympart=sympart,
                                  show_plots=show_plots,
                                  rows=3,  # rows
                                  cols=2,  # cols
                                  num=3)   # start num

        if da > 0.0 or dy > 0.0:
            if inner:
                r_out = da / 2.0
                r_in = dy / 2.0
            elif outer:
                r_out = dy / 2.0
                r_in = da / 2.0
            else:
                r_out = 0.0
                r_in = 0.0
            if machine.cut_is_possible(r_in, r_out):
                logger.info("make a cut")
                machine = machine.cut(r_in, r_out)

        if part:
            if part[0] == 'stator':
                machine.geom.set_stator()
                machine.geom.search_stator_subregions(part[1])

                if machine.has_mirrored_windings():
                    logger.info("undo mirror of stator")
                    machine = machine.undo_mirror()
                    machine.geom.set_stator()
                    machine.geom.search_stator_subregions(part[1])
                    machine.create_mirror_lines_outside_windings()

                params = create_femag_parameters_stator(machine,
                                                        part[1])
            else:
                machine.geom.set_rotor()
                machine.geom.search_rotor_subregions(part[1])
                params = create_femag_parameters_rotor(machine,
                                                       part[1])
        else:
            machine.geom.search_subregions()

        machine.delete_tiny_elements(mindist)
        logger.info("END of work: %s", basename)

        if show_plots:
            p.render_elements(machine.geom, Shape,
                              draw_inside=True, title=name,
                              rows=3, cols=2, num=5, show=False,
                              # with_nodes=True,
                              # neighbors=True,
                              write_id=write_id,
                              fill_areas=True)
            if write_png:
                p.write_plot(basename)
            else:
                p.show_plot()

        if show_areas:
            p.render_elements(machine.geom, Shape,
                              title=name,
                              show=True,
                              draw_inside=True,
                              neighbors=True,
                              fill_areas=True)
            p.render_areas(machine.geom,
                           title=name,
                           with_nodes=True,
                           single_view=True)

        if write_fsl:
            if machine.is_full():
                logger.warning("it's not possible to create fsl-file")
                return None

            fslrenderer = FslRenderer(basename)
            conv['fsl'] = fslrenderer.render(machine, inner, outer)
            if params:
                conv.update(params)

    conv['name'] = basename
    return conv


def create_femag_parameters(m_inner, m_outer, nodedist=1):
    if not (m_inner and m_outer):
        return {}

    params = {}
    geom_inner = m_inner.geom
    geom_outer = m_outer.geom

    parts_inner = int(m_inner.get_symmetry_part())
    parts_outer = int(m_outer.get_symmetry_part())

    if parts_inner > parts_outer:
        num_slots = parts_inner
        num_poles = parts_outer
        num_sl_gen = int(geom_inner.get_symmetry_copies()+1)
        alfa_slot = geom_inner.get_alfa()
        alfa_pole = geom_outer.get_alfa()
    else:
        num_slots = parts_outer
        num_poles = parts_inner
        num_sl_gen = int(geom_outer.get_symmetry_copies()+1)
        alfa_slot = geom_outer.get_alfa()
        alfa_pole = geom_inner.get_alfa()

    params['tot_num_slot'] = num_slots
    params['num_sl_gen'] = num_sl_gen
    params['num_poles'] = num_poles
    params['nodedist'] = nodedist

    params['dy1'] = 2*geom_outer.max_radius
    params['da1'] = 2*geom_outer.min_radius
    params['da2'] = 2*geom_inner.max_radius
    params['dy2'] = 2*geom_inner.min_radius
    params['agndst'] = agndst(params['da1'], params['da2'],
                              num_slots, num_poles,
                              nodedist)
    params['alfa_slot'] = alfa_slot
    params['alfa_pole'] = alfa_pole

    if num_slots == 0 or num_poles == 0:
        if num_slots == 0:
            logger.warning("No slots found")
        if num_poles == 0:
            logger.warning("No poles found")
        logger.warning("Model not ready for femag")
        return {'error': "Model not ready for femag"}

    if not np.isclose(alfa_slot * num_slots,
                      alfa_pole * num_poles):
        logger.warning("slots and poles dont match")
        return {'error': "Model not ready for femag"}

    return params


def create_femag_parameters_stator(motor, position):
    params = {}
    num_slots = motor.get_num_slots()
    params['tot_num_slot'] = num_slots
    if position == 'in':
        params['da2'] = 2*motor.geom.max_radius
        params['dy2'] = 2*motor.geom.min_radius
    else:
        params['dy1'] = 2*motor.geom.max_radius
        params['da1'] = 2*motor.geom.min_radius
    return params


def create_femag_parameters_rotor(motor, position):
    params = {}
    num_poles = motor.get_num_poles()
    params['num_poles'] = num_poles
    if position == 'in':
        params['da2'] = 2*motor.geom.max_radius
        params['dy2'] = 2*motor.geom.min_radius
    else:
        params['dy1'] = 2*motor.geom.max_radius
        params['da1'] = 2*motor.geom.min_radius
    return params



"""Netlist creation """

def create(num_poles, model_divisor, voltage_source, rotor_cage):
    rotor_slots_gen = rotor_cage['num_slots'] // model_divisor
    num_poles_gen = num_poles//model_divisor
    NoPhases = len(voltage_source['phiq'])
    StatorNodes=2*NoPhases+1

    uq = [dict(
        Type = "Uac",
        Name = f"Uq{i+1}",
        Nodes = [i+1, 0],
        Amplitude = voltage_source['Uq_amp'],
        Frequency = voltage_source['fq'],
        Phaseangle = phiq,
        Ri = voltage_source['Rqi'])
              for i, phiq in enumerate(voltage_source['phiq'])]

    hf = [dict(
        Type = "Uac",
        Name = f"Uq{i+1}_HF",
        Nodes = [i+1,i+NoPhases+1],
        Amplitude = 0,
        Frequency = 0,
        Phaseangle = 0,
        Ri = 0) for i in range(NoPhases)]

    stator_wdg = [dict(
        Type = "Wdg",
        Name = f"PH{i+1}",
        Nodes = [i+NoPhases+1,StatorNodes],
        WdgKey = i+1) for i in range(NoPhases)]
    
    rotor_bars = [dict(
        Type = "Wdg",
        Name = f"Bar{i+1}",
        Nodes = [i+StatorNodes, i+StatorNodes+rotor_slots_gen],
        WdgKey = i+1+NoPhases ) for i in range(rotor_slots_gen)]
    
    rotor_bars[0]['Nodes'] = [0, StatorNodes+rotor_slots_gen] 

    rotor_ring_segments = [dict(
        Type = "RL",
        Name = f"Z{i+1}",
        Nodes = [i+StatorNodes, i+StatorNodes+1],
        R = rotor_cage['RR']*model_divisor,
        L = rotor_cage['LR']*model_divisor)
                               for i in range(rotor_slots_gen*2)]
        
    rotor_ring_segments[0]['Nodes'] = [0,StatorNodes+1]
    if num_poles_gen % 2 == 0:   # even number of poles in model
        i = rotor_slots_gen-1
        rotor_ring_segments[i]['Nodes'] = [i+StatorNodes, 0]
        i = 2*rotor_slots_gen-1
        rotor_ring_segments[i]['Nodes'] = [i+StatorNodes, StatorNodes + rotor_slots_gen]
    else:
        i = 2*rotor_slots_gen-1
        rotor_ring_segments[i]['Nodes'] = [i+StatorNodes, 0]

    return dict(
        Branches=uq+hf+stator_wdg+rotor_bars+rotor_ring_segments )
